#  -------------------------------------------------------------------------
#  pyCGNS - Python package for CFD General Notation System -
#  See license.txt file in the root directory of this Python module source
#  -------------------------------------------------------------------------
#
# TESTING APP
#
import unittest

import CGNS.PAT.cgnslib as CGL
import CGNS.PAT.cgnsutils as CGU
import CGNS.PAT.cgnskeywords as CGK
import numpy as NPY

import importlib
import os
import string
import subprocess


def genTrees():
    T = CGL.newCGNSTree()
    b = CGL.newBase(T, "{Base}", 2, 3)
    z = CGL.newZone(b, "{Zone}", NPY.array([[5, 4, 0], [7, 6, 0]], order="F"))
    g = CGL.newGridCoordinates(z, "GridCoordinates")
    d = CGL.newDataArray(g, CGK.CoordinateX_s, NPY.ones((5, 7), dtype="d", order="F"))
    d = CGL.newDataArray(g, CGK.CoordinateY_s, NPY.ones((5, 7), dtype="d", order="F"))
    d = CGL.newDataArray(g, CGK.CoordinateZ_s, NPY.ones((5, 7), dtype="d", order="F"))
    return (T,)


class APPTestCase(unittest.TestCase):
    def setUp(self):
        pass

    def test_000_Module(self):
        import CGNS.APP.lib


# ---
print("-" * 70 + "\nCGNS.APP test suite")
suite = unittest.TestLoader().loadTestsFromTestCase(APPTestCase)
unittest.TextTestRunner(verbosity=2).run(suite)

# --- last line



#  -------------------------------------------------------------------------
#  pyCGNS.VAL - Python package for CFD General Notation System -
#  See license.txt file in the root directory of this Python module source
#  -------------------------------------------------------------------------
#
import CGNS.VAL.grammars.CGNS_VAL_USER_DEFAULT as CGV
import CGNS.VAL.parse.messages as CGM
import CGNS.PAT.cgnserrors as CGE
import CGNS.VAL.parse.findgrammar

import sys
import os


def listuserkeys(trace):
    s = ""
    if not trace:
        print("### Use -v (verbose) option to check if the found paths")
        print("### are into your PYTHONPATH. If not, grammar would not be")
        print("### used by CGNS.VAL even if detected...")
    dk = CGNS.VAL.parse.findgrammar.findAllUserGrammars(trace)
    for key in dk:
        s += "%-16s: %s\n" % (key, dk[key])
    return s


def getParser(trace, user):
    if user is not None:
        mod = CGNS.VAL.parse.findgrammar.importUserGrammars(user, verbose=trace)
        if mod is None:
            if trace:
                print("### Using grammar [DEFAULT]")
            parser = CGV.CGNS_VAL_USER_Checks(None)
        else:
            if trace:
                print("### Using grammar [%s]" % user)
            if trace:
                print("### Found in [%s]" % os.path.dirname(mod.__file__))
            parser = mod.CGNS_VAL_USER_Checks(None)
    else:
        parser = CGV.CGNS_VAL_USER_Checks(None)
    return parser


def listdiags(trace, userlist):
    ld = []
    for user in userlist:
        parser = getParser(trace, user)
        mlist = parser.listDiagnostics()
        ld += list(mlist)
    ld.sort()
    s = ""
    for d in ld:
        s += "%s\n" % (str(mlist[d]))
    return s


def run(T, trace, userlist, stop=False, warnings=[], failures=[]):
    diag = CGM.DiagnosticLog()
    for user in userlist:
        parser = getParser(trace, user)
        # parser lookup and init are required prior to initialize message tables
        for w in warnings:
            diag.forceAsWarning(w)
        for w in failures:
            diag.forceAsFailure(w)
        try:
            parser.checkTree(T, trace, stop=stop)
        except (CGE.cgnsException,) as v:
            pass
        diag.merge(parser.log)
    return diag


def compliant(
    T,
    trace=False,
    userlist=["DEFAULT"],
    paths=[""],
    stop=False,
    warnings=[],
    failures=[],
):
    ipath = "%s/lib/python%s.%s/site-packages/CGNS/VAL/grammars" % (
        sys.prefix,
        sys.version_info[0],
        sys.version_info[1],
    )
    sys.path.append(ipath)
    for pp in paths:
        sys.path.append(pp)
    diag = run(T, trace, userlist, stop=stop, warnings=warnings, failures=failures)
    ok = [True, []]
    for p in diag:
        for (s, sp) in diag.diagnosticsByPath(p):
            if diag.status(s) == CGM.CHECK_FAIL:
                ok[0] = False
                ok[1].append((p, diag.message(s)))
    return ok


def showDiag(diag, idlist, bypath=True):
    ok = True
    if bypath:
        for p in diag:
            if not diag.hasOnlyKey(p, idlist):
                print("\n%s\n%s" % ("-" * 75, p))
                for (s, sp) in diag.diagnosticsByPath(p):
                    if (diag.status(s) != CGM.CHECK_GOOD) and (
                        diag.key(s) not in idlist
                    ):
                        print(diag.message(s))
                        if diag.status(s) == CGM.CHECK_FAIL:
                            ok = False
        print("\n%s\n" % ("-" * 75))
    else:
        for m in diag.allMessageKeys():
            if m not in idlist:
                first = True
                ctxt = diag.noContextMessage(m)
                if ctxt is not None:
                    print("\n%s\n[%s] %s" % ("-" * 75, m, ctxt))
                else:
                    print("\n%s\n[%s]" % ("-" * 75, m))
                for (d, dp) in diag.diagnosticsByMessage(m):
                    if diag.status(d) != CGM.CHECK_GOOD:
                        if ctxt is None:
                            if not first:
                                skip = "\n"
                            else:
                                skip = ""
                            print("%s  %s\n  > %s" % (skip, dp, d.message))
                        else:
                            print("  %s" % (dp))
                        first = False
                        if diag.status(d) == CGM.CHECK_FAIL:
                            ok = False
        print("\n%s" % ("-" * 75))
    if ok:
        print("### CGNS/Python tree Compliant")
    else:
        print("### CGNS/Python tree *NOT* Compliant")



#  -------------------------------------------------------------------------
#  pyCGNS - Python package for CFD General Notation System -
#  See license.txt file in the root directory of this Python module source
#  -------------------------------------------------------------------------
#

from CGNS.NAV.moption import Q7OptionContext as OCTXT

import sys
import string
import re

from qtpy.QtWidgets import QTextEdit, QDialog
from qtpy.QtGui import QTextCursor, QSyntaxHighlighter, QTextCharFormat, QFont, QColor


from CGNS.NAV.Q7MessageWindow import Ui_Q7MessageWindow
from CGNS.NAV.Q7LogWindow import Ui_Q7LogWindow

(INFO, QUESTION, ERROR, WARNING) = (0, 1, 2, 3)

# globals
OK = True
CANCEL = False
ANSWER = True


# -----------------------------------------------------------------
class Q7DocEditor(QTextEdit):
    def __init__(self, parent=None):
        QTextEdit.__init__(self, parent)

    def initText(self, text):
        self.clear()
        self.insertPlainText(text)
        self.moveCursor(QTextCursor.Start, QTextCursor.MoveAnchor)
        self.ensureCursorVisible()


# -----------------------------------------------------------------
class Q7PythonEditor(QTextEdit):
    def __init__(self, parent=None):
        QTextEdit.__init__(self, parent)
        self.highlighter = Q7PythonEditorHighlighter(self.document())
        fsz = 12
        ffm = "Courier"
        self.setStyleSheet('font: %dpt "%s";' % (fsz, ffm))

    def initText(self, text):
        self.clear()
        self.insertPlainText(text)
        self.moveCursor(QTextCursor.Start, QTextCursor.MoveAnchor)
        self.ensureCursorVisible()


# -----------------------------------------------------------------
class Q7PythonEditorHighlighter(QSyntaxHighlighter):
    def __init__(self, *args):
        keywords = r"\bimport\b|\bas\b|\bfor\b|\bdef\b|\bpass\b"
        keywords += r"|\bclass\b|\bfrom\b|\bif\b|\bthen\b|\belse\b"
        keywords += r"|\belif\b|\btry\b|\bexcept\b|\bfinally\b|\braise\b"
        keywords += r"|\bprint\b|\bin\b|\bnot\b|\band\b|\bor\b|\bcontinue\b"
        keywords += r"|\bwhile\b|\breturn\b|\blambda\b|\bwith\b|\del\b"
        keywords += r"|\bglobal\b|\byield\b|\bexec\b|\bassert\b|\break\b"
        constants = r"\bNone\b|\bTrue\b|\bFalse\b|\bself\b"
        autovars = r"\bNODE\b|\bNAME\b|\bVALUE\b|\bSIDSTYPE\b|\bCHILDREN\b"
        autovars += r"|\bTREE\b|\bPATH\b|\bRESULT\b|\bARGS\b|\bPARENT\b"
        autovars += r"|\bLINKS\b|\bSKIPS\b"
        numbers = r"[-+]?\d+" + "|[-+]?(\d+(\.\d*)?|\.\d+)([eE][-+]?\d+)?"
        comment = r"\#.*$"
        textstring = r'"[^"]*?"|\'[^\']*?\''
        self.f_keywords = QTextCharFormat()
        self.f_keywords.setFontWeight(QFont.Bold)
        self.f_keywords.setForeground(QColor("Blue"))
        self.f_comment = QTextCharFormat()
        self.f_comment.setFontWeight(QFont.Light)
        self.f_comment.setForeground(QColor("Red"))
        self.f_constants = QTextCharFormat()
        self.f_constants.setFontWeight(QFont.Light)
        self.f_constants.setForeground(QColor("Navy"))
        self.f_autovars = QTextCharFormat()
        self.f_autovars.setFontWeight(QFont.Bold)
        self.f_autovars.setForeground(QColor("Green"))
        self.f_textstring = QTextCharFormat()
        self.f_textstring.setFontWeight(QFont.Bold)
        self.f_textstring.setForeground(QColor("Coral"))
        self.f_numbers = QTextCharFormat()
        self.f_numbers.setFontWeight(QFont.Light)
        self.f_numbers.setForeground(QColor("Gray"))
        self.r_keywords = re.compile(keywords)
        self.r_numbers = re.compile(numbers)
        self.r_comment = re.compile(comment)
        self.r_constants = re.compile(constants)
        self.r_autovars = re.compile(autovars)
        self.r_textstring = re.compile(textstring)
        self.in_comment = False
        self.in_string = False
        self.r_syntax = []
        self.r_syntax.append((self.r_keywords, self.f_keywords))
        self.r_syntax.append((self.r_numbers, self.f_numbers))
        self.r_syntax.append((self.r_autovars, self.f_autovars))
        self.r_syntax.append((self.r_constants, self.f_constants))
        self.r_syntax.append((self.r_comment, self.f_comment))
        self.r_syntax.append((self.r_textstring, self.f_textstring))
        QSyntaxHighlighter.__init__(self, *args)

    def highlightBlock(self, textblock):
        text = str(textblock)
        fdef = self.format(0)
        fmax = len(text)
        for (rex, fmt) in self.r_syntax:
            mit = rex.finditer(text)
            for m in mit:
                self.setFormat(m.span()[0], m.span()[1], fmt)
                self.setFormat(m.span()[1], fmax, fdef)


# -----------------------------------------------------------------
class Q7Log(QDialog, Ui_Q7LogWindow):
    def __init__(self):
        QDialog.__init__(self, None)
        self.setupUi(self)
        self.bClose.clicked.connect(self.leave)
        self.bClear.clicked.connect(self.clear)
        self.setWindowTitle("%s: Log" % OCTXT._ToolName)
        self.eLog.setReadOnly(True)
        self.eLog.setAcceptRichText(False)
        self.eLog.setStyleSheet('font: 12pt "Courier";')

    def write(self, msg):
        self.eLog.insertPlainText(msg)

    def clear(self):
        self.eLog.clear()

    def leave(self, *arg):
        self.hide()


# -----------------------------------------------------------------
class Q7MessageBox(QDialog, Ui_Q7MessageWindow):
    def __init__(self, control, code):
        QDialog.__init__(self, None)
        self.setupUi(self)
        self.bOK.clicked.connect(self.runOK)
        self.bCANCEL.clicked.connect(self.runCANCEL)
        self.bInfo.setDisabled(True)
        self._text = ""
        self._control = control
        self._code = 0

    def setMode(self, cancel=False, again=False):
        if not again:
            self.cNotAgain.hide()
        if not cancel:
            self.bCANCEL.hide()

    def setLayout(
        self, text, btype=INFO, cancel=False, again=False, buttons=("OK", "Cancel")
    ):
        self.bOK.setText(buttons[0])
        if len(buttons) > 1:
            self.bCANCEL.setText(buttons[1])
        self.eMessage.setText(str(text))
        self._text = text
        self.eMessage.setReadOnly(True)
        self.setMode(cancel, again)

    def runOK(self, *arg):
        self.addToSkip()
        self.close()
        self.setResult(QDialog.Accepted)

    def runCANCEL(self, *arg):
        self.addToSkip()
        self.close()
        self.setResult(QDialog.Rejected)

    def addToSkip(self):
        if self.cNotAgain.isChecked():
            if self._code and self._code not in OCTXT.IgnoredMessages:
                OCTXT.IgnoredMessages += [self._code]

    def showAndWait(self):
        ret = self.exec_()
        if ret == QDialog.Accepted:
            return True
        else:
            return False


def wError(control, code, info, error):
    txt = """<img source=":/images/icons/user-G.png">  <big>ERROR #%d</big><hr>
         %s<br>%s""" % (
        code,
        error,
        info,
    )
    if code in OCTXT.IgnoredMessages:
        return True
    msg = Q7MessageBox(control, code)
    msg.setWindowTitle("%s: Error" % OCTXT._ToolName)
    msg.setLayout(txt, btype=ERROR, cancel=False, again=True, buttons=("Close",))
    return msg.showAndWait()


def wQuestion(control, code, title, question, again=True, buttons=("OK", "Cancel")):
    txt = """<img source=":/images/icons/user-M.png">
         <b> <big>%s</big></b><hr>%s""" % (
        title,
        question,
    )
    if code in OCTXT.IgnoredMessages:
        return True
    msg = Q7MessageBox(control, code)
    msg.setWindowTitle("%s: Question" % OCTXT._ToolName)
    msg.setLayout(txt, btype=QUESTION, cancel=True, again=again, buttons=buttons)
    return msg.showAndWait()


def wInfo(control, code, title, info, again=True, buttons=("Close",)):
    txt = """<img source=":/images/icons/user-S.png">
         <b> <big>%s</big></b><hr>%s""" % (
        title,
        info,
    )
    if code in OCTXT.IgnoredMessages:
        return True
    msg = Q7MessageBox(control, code)
    msg.setWindowTitle("%s: Info" % OCTXT._ToolName)
    msg.setLayout(txt, btype=INFO, cancel=False, again=again, buttons=buttons)
    return msg.showAndWait()


# --- last line



#  -------------------------------------------------------------------------
#  pyCGNS.NAV - Python package for CFD General Notation System - NAVigater
#  See license.txt file in the root directory of this Python module source
#  -------------------------------------------------------------------------
#

from CGNS.NAV.moption import Q7OptionContext as OCTXT

import CGNS.PAT.cgnsutils as CGU

from qtpy.QtCore import Qt, QModelIndex, QPoint
from qtpy.QtWidgets import QStyledItemDelegate, QMenu
from qtpy.QtGui import QColor, QFont

from CGNS.NAV.Q7DiffWindow import Ui_Q7DiffWindow
from CGNS.NAV.wfingerprint import Q7Window, Q7FingerPrint
from CGNS.NAV.mdifftreeview import Q7DiffTreeModel

import CGNS.NAV.mtree as NMT

(DIFF_NX, DIFF_NA, DIFF_ND, DIFF_CQ, DIFF_CT, DIFF_CS, DIFF_CV) = range(7)


# -----------------------------------------------------------------
class Q7DiffItemDelegate(QStyledItemDelegate):
    def __init__(self, owner, model, diag):
        QStyledItemDelegate.__init__(self, owner)
        self._parent = owner
        self._model = model
        self._diag = diag
        self.orange = QColor("#FFA500")
        self.red = QColor("#FF0000")
        self.green = QColor("#008000")
        self.gray = QColor("#C0C0C0")

    def paint(self, painter, option, index):
        idx = self._parent.modelIndex(index)
        col = idx.column()
        nnm = self._parent.modelData(index).sidsName()
        pth = CGU.getPathNoRoot(self._parent.modelData(index).sidsPath())
        if col == NMT.COLUMN_NAME:
            if nnm not in OCTXT._ReservedNames:
                option.font.setWeight(QFont.Bold)
            QStyledItemDelegate.paint(self, painter, option, index)
            option.font.setWeight(QFont.Light)
            if (self._diag is not None) and (pth in self._diag):
                color = self.gray
                if self._diag[pth] == DIFF_NX:
                    color = self.gray
                if self._diag[pth] == DIFF_NA:
                    color = self.green
                if self._diag[pth] == DIFF_ND:
                    color = self.red
                painter.save()
                painter.setBrush(Qt.NoBrush)
                painter.setPen(color)
                painter.drawRect(option.rect)
                painter.restore()
        elif col == NMT.COLUMN_VALUE:
            option.font.setFamily(OCTXT.Table_Family)
            QStyledItemDelegate.paint(self, painter, option, index)
            if (self._diag is not None) and (pth in self._diag):
                if self._diag[pth] == DIFF_CV:
                    painter.save()
                    painter.setBrush(Qt.NoBrush)
                    painter.setPen(self.orange)
                    painter.drawRect(option.rect)
                    painter.restore()
        elif col == NMT.COLUMN_DATATYPE:
            option.font.setFamily(OCTXT.Table_Family)
            option.font.setPointSize(8)
            QStyledItemDelegate.paint(self, painter, option, index)
            if (self._diag is not None) and (pth in self._diag):
                if self._diag[pth] == DIFF_CQ:
                    painter.save()
                    painter.setBrush(Qt.NoBrush)
                    painter.setPen(self.orange)
                    painter.drawRect(option.rect)
                    painter.restore()
        elif col == NMT.COLUMN_SHAPE:
            QStyledItemDelegate.paint(self, painter, option, index)
            if (self._diag is not None) and (pth in self._diag):
                if self._diag[pth] == DIFF_CS:
                    painter.save()
                    painter.setBrush(Qt.NoBrush)
                    painter.setPen(self.orange)
                    painter.drawRect(option.rect)
                    painter.restore()
        elif col == NMT.COLUMN_SIDS:
            QStyledItemDelegate.paint(self, painter, option, index)
            if (self._diag is not None) and (pth in self._diag):
                if self._diag[pth] == DIFF_CT:
                    painter.save()
                    painter.setBrush(Qt.NoBrush)
                    painter.setPen(self.orange)
                    painter.drawRect(option.rect)
                    painter.restore()
        else:
            QStyledItemDelegate.paint(self, painter, option, index)


# -----------------------------------------------------------------
class Q7Diff(Q7Window, Ui_Q7DiffWindow):
    def __init__(self, control, fgprintindexA, fgprintindexB, diag):
        Q7Window.__init__(self, Q7Window.VIEW_DIFF, control, None, fgprintindexA)
        self._depthExpanded = 0
        self._lastEntered = None
        self._fgidxA = fgprintindexA
        self._fgidxB = fgprintindexB
        ldiag = self.diagAnalysis(diag)
        self.treeviewA.expanded[QModelIndex].connect(self.expandNode)
        self.treeviewA.collapsed.connect(self.collapseNode)
        self.treeviewA.clicked[QModelIndex].connect(self.clickedNode)
        self.treeviewA.customContextMenuRequested[QPoint].connect(self.clickedNode)
        self.treeviewB.expanded[QModelIndex].connect(self.expandNode)
        self.treeviewB.collapsed.connect(self.collapseNode)
        self.treeviewB.clicked[QModelIndex].connect(self.clickedNode)
        self.treeviewB.customContextMenuRequested[QPoint].connect(self.clickedNode)
        # QObject.connect(self.treeviewA,
        #                 SIGNAL("expanded(QModelIndex)"),
        #                 self.expandNode)
        # QObject.connect(self.treeviewA,
        #                 SIGNAL("collapsed()"),
        #                 self.collapseNode)
        # QObject.connect(self.treeviewA,
        #                 SIGNAL("clicked(QModelIndex)"),
        #                 self.clickedNode)
        # QObject.connect(self.treeviewA,
        #                 SIGNAL("customContextMenuRequested(QPoint)"),
        #                 self.clickedNode)
        # QObject.connect(self.treeviewB,
        #                 SIGNAL("expanded(QModelIndex)"),
        #                 self.expandNode)
        # QObject.connect(self.treeviewB,
        #                 SIGNAL("collapsed()"),
        #                 self.collapseNode)
        # QObject.connect(self.treeviewB,
        #                 SIGNAL("clicked(QModelIndex)"),
        #                 self.clickedNode)
        # QObject.connect(self.treeviewB,
        #                 SIGNAL("customContextMenuRequested(QPoint)"),
        #                 self.clickedNode)
        # QObject.connect(self.bLockScroll,
        #                 SIGNAL("clicked()"),
        #                 self.syncScrolls)
        self.bLockScroll.clicked.connect(self.syncScrolls)
        self.bClose.clicked.connect(self.leave)
        self.bInfo.clicked.connect(self.infoTreeView)
        self.bZoomIn.clicked.connect(self.expandLevel)
        self.bZoomOut.clicked.connect(self.collapseLevel)
        self.bZoomAll.clicked.connect(self.expandMinMax)
        self.bSaveDiff.clicked.connect(self.savediff)
        self.setContextMenuPolicy(Qt.CustomContextMenu)
        self.popupmenu = QMenu()
        self.proxyA = Q7DiffTreeModel(self._fgidxA)
        self.proxyB = Q7DiffTreeModel(self._fgidxB)
        self.proxyA.setDiag(ldiag)
        self.proxyB.setDiag(ldiag)
        self.treeviewA.setModel(self.proxyA)
        self.treeviewB.setModel(self.proxyB)
        fga = Q7FingerPrint.getByIndex(self._fgidxA)
        fgb = Q7FingerPrint.getByIndex(self._fgidxB)
        model_a = fga.model
        model_b = fgb.model
        self.treeviewA.setItemDelegate(
            Q7DiffItemDelegate(self.treeviewA, model_a, ldiag)
        )
        self.treeviewB.setItemDelegate(
            Q7DiffItemDelegate(self.treeviewB, model_b, ldiag)
        )
        self.treeviewA.setControlWindow(self, model_a)
        self.treeviewB.setControlWindow(self, model_b)
        self.treeviewA.hideColumn(NMT.COLUMN_FLAG_LINK)
        self.treeviewA.hideColumn(NMT.COLUMN_FLAG_CHECK)
        self.treeviewA.hideColumn(NMT.COLUMN_FLAG_SELECT)
        self.treeviewA.hideColumn(NMT.COLUMN_FLAG_USER)
        self.treeviewB.hideColumn(NMT.COLUMN_FLAG_LINK)
        self.treeviewB.hideColumn(NMT.COLUMN_FLAG_CHECK)
        self.treeviewB.hideColumn(NMT.COLUMN_FLAG_SELECT)
        self.treeviewB.hideColumn(NMT.COLUMN_FLAG_USER)
        self.wvsa = self.treeviewA.verticalScrollBar()
        self.wvsb = self.treeviewB.verticalScrollBar()
        self.uvsa = self.verticalScrollBarA
        self.uvsb = self.verticalScrollBarB
        self.uvsa.setToolTip("%s/%s" % (fga.filedir, fga.filename))
        self.uvsb.setToolTip("%s/%s" % (fgb.filedir, fgb.filename))
        self.syncScrolls(True)

    def diagAnalysis(self, diag):
        ldiag = {}
        for k in diag:
            ldiag[k] = DIFF_NX
            for d in diag[k]:
                if d[0] == "NA":
                    ldiag[d[1]] = DIFF_NA
                if d[0] == "ND":
                    ldiag[d[1]] = DIFF_ND
                if d[0] in ["CT"]:
                    ldiag[k] = DIFF_CT
                if d[0] in ["C3", "C1", "C2"]:
                    ldiag[k] = DIFF_CQ
                if d[0] in ["C6", "C7"]:
                    ldiag[k] = DIFF_CV
                if d[0] in ["C4", "C5"]:
                    ldiag[k] = DIFF_CS
        return ldiag

    def syncScrolls(self, force=False):
        self.uvsa.valueChanged[int].connect(self.wvsa.setValue[int])
        self.uvsb.valueChanged[int].connect(self.wvsb.setValue[int])
        self.wvsa.valueChanged[int].connect(self.uvsa.setValue[int])
        self.wvsb.valueChanged[int].connect(self.uvsb.setValue[int])
        #
        # QObject.connect(self.uvsa, SIGNAL("valueChanged(int)"),
        #                 self.wvsa, SLOT("setValue(int)"))
        # QObject.connect(self.uvsb, SIGNAL("valueChanged(int)"),
        #                 self.wvsb, SLOT("setValue(int)"))
        # QObject.connect(self.wvsa, SIGNAL("valueChanged(int)"),
        #                 self.uvsa, SLOT("setValue(int)"))
        # QObject.connect(self.wvsb, SIGNAL("valueChanged(int)"),
        #                 self.uvsb, SLOT("setValue(int)"))
        if force or self.bLockScroll.isChecked():
            self.wvsa.valueChanged[int].connect(self.wvsb.setValue[int])
            self.wvsb.valueChanged[int].connect(self.wvsa.setValue[int])
            # QObject.connect(self.wvsa, SIGNAL("valueChanged(int)"),
            #                 self.wvsb, SLOT("setValue(int)"))
            # QObject.connect(self.wvsb, SIGNAL("valueChanged(int)"),
            #                 self.wvsa, SLOT("setValue(int)"))
        else:
            self.wvsa.valueChanged[int].disconnect(self.wvsb.setValue[int])
            self.wvsb.valueChanged[int].disconnect(self.wvsa.setValue[int])
            # QObject.disconnect(self.wvsa, SIGNAL("valueChanged(int)"),
            #                   self.wvsb, SLOT("setValue(int)"))
            # QObject.disconnect(self.wvsb, SIGNAL("valueChanged(int)"),
            #                   self.wvsa, SLOT("setValue(int)"))

    def model(self):
        return self._fgprint.model

    def modelIndex(self, idx):
        if not idx.isValid():
            return -1
        midx = idx
        # if (idx.model() != self.treeview.M()):
        #    midx=self.treeview.model().mapToSource(idx)
        return midx

    def modelData(self, idx):
        if not idx.isValid():
            return None
        return self.modelIndex(idx).internalPointer()

    def infoTreeView(self):
        self._control.helpWindow("Tree")

    def savediff(self):
        pass

    def expandMinMax(self):
        if self._depthExpanded == Q7FingerPrint.getByIndex(self._fgidxA).depth - 2:
            self._depthExpanded = -1
            self.treeviewA.collapseAll()
            self.treeviewB.collapseAll()
        else:
            self._depthExpanded = Q7FingerPrint.getByIndex(self._fgidxA).depth - 2
            self.treeviewA.expandAll()
            self.treeviewB.expandAll()
        self.resizeAll()

    def expandLevel(self):
        if self._depthExpanded < Q7FingerPrint.getByIndex(self._fgidxA).depth - 2:
            self._depthExpanded += 1
        self.treeviewA.expandToDepth(self._depthExpanded)
        self.treeviewB.expandToDepth(self._depthExpanded)
        self.resizeAll()

    def collapseLevel(self):
        if self._depthExpanded != -1:
            self._depthExpanded -= 1
        if self._depthExpanded == -1:
            self.treeviewA.collapseAll()
            self.treeviewB.collapseAll()
        else:
            self.treeviewA.expandToDepth(self._depthExpanded)
            self.treeviewB.expandToDepth(self._depthExpanded)
        self.resizeAll()

    def updateStatus(self, node):
        return
        self.lineEdit.clear()
        self.lineEdit.insert(node.sidsPath())

    def updateMenu(self, nodeidxs):
        return
        nodeidx = self.modelIndex(nodeidxs)
        if not nodeidx.isValid:
            return False
        if nodeidx.internalPointer() is None:
            return False
        if nodeidx.internalPointer().sidsPath() == "/CGNSTree":
            return False
        self.setLastEntered(nodeidxs)
        if nodeidx != -1:
            node = nodeidx.internalPointer()
            lknode = not node.sidsIsLink()
            lznode = node.hasLazyLoad()
            actlist = (
                ("About %s" % node.sidsType(), self.aboutSIDS, None, False),
                None,
                ("Mark/unmark node", self.marknode, "Space", False),
                ("Add new child node", self.newnodechild, "Ctrl+A", False),
                ("Add new brother node", self.newnodebrother, "Ctrl+Z", False),
                None,
                ("Open form", self.popform, "Ctrl+F", False),
                ("Open view", self.openSubTree, "Ctrl+W", False),
                ("Open view on linked-to file", self.openLkTree, "Ctrl+O", lknode),
                None,
                ("Load node data in memory", self.dataLoad, "Ctrl+L", not lznode),
                ("Release memory node data", self.dataRelease, "Ctrl+R", lznode),
                None,
                ("Copy", self.mcopy, "Ctrl+C", False),
                ("Cut", self.mcut, "Ctrl+X", False),
                ("Paste as brother", self.mpasteasbrother, "Ctrl+V", False),
                ("Paste as child", self.mpasteaschild, "Ctrl+Y", False),
                None,
                ("Cut all selected", self.mcutselected, "Ctrl+Shift+X", False),
                (
                    "Paste as brother for each selected",
                    self.mpasteasbrotherselected,
                    "Ctrl+Shift+V",
                    False,
                ),
                (
                    "Paste as child for each selected",
                    self.mpasteaschildselected,
                    "Ctrl+Shift+Y",
                    False,
                ),
                (
                    "Load nodes data in memory for each selected",
                    self.dataLoadSelected,
                    "Ctrl+Shift+L",
                    False,
                ),
                (
                    "Release memory node data for each selected",
                    self.dataReleaseSelected,
                    "Ctrl+Shift+R",
                    False,
                ),
            )
            self.popupmenu.clear()
            self.popupmenu.setTitle("Node menu")
            for aparam in actlist:
                if aparam is None:
                    self.popupmenu.addSeparator()
                else:
                    a = QAction(aparam[0], self)
                    a.triggered.connect(aparam[1])
                    if aparam[2] is not None:
                        a.setShortcut(aparam[2])
                    self.popupmenu.addAction(a)
                    a.setDisabled(aparam[3])
            return True

    def setLastEntered(self, nix=None):
        self._lastEntered = None

    def getLastEntered(self):
        return self._lastEntered

    def clearLastEntered(self):
        self._lastEntered = None
        self.treeview.selectionModel().clearSelection()
        return None

    def clickedNode(self, index):
        pass

    def expandNode(self, *args):
        self.resizeAll()

    def collapseNode(self, *args):
        pass

    def resizeAll(self):
        for n in range(NMT.COLUMN_LAST + 1):
            self.treeviewA.resizeColumnToContents(n)
            self.treeviewB.resizeColumnToContents(n)

    def show(self):
        super(Q7Diff, self).show()

    def closeAlone(self):
        pass

    def leave(self):
        self.close()

    def doRelease(self):
        pass


# -----------------------------------------------------------------



#  -------------------------------------------------------------------------
#  pyCGNS.VAL - Python package for CFD General Notation System - VALidater
#  See license.txt file in the root directory of this Python module source
#  -------------------------------------------------------------------------
#
import CGNS.PAT.cgnslib as CGL
import CGNS.PAT.cgnsutils as CGU
import CGNS.PAT.cgnskeywords as CGK
import numpy as NPY

TESTS = []

#  -------------------------------------------------------------------------
tag = "gridconnectivity"
diag = True
vertexsize = 20
cellsize = 7
ntris = 12


def makeCorrectTree():
    T = CGL.newCGNSTree()
    b = CGL.newBase(T, "Base", 3, 3)
    z1 = CGL.newZone(
        b, "Zone1", NPY.array([[5, 4, 0], [7, 6, 0], [9, 8, 0]], order="F")
    )
    g = CGL.newGridCoordinates(z1, "GridCoordinates")
    d = CGL.newDataArray(
        g, CGK.CoordinateX_s, NPY.ones((5, 7, 9), dtype="float64", order="F")
    )
    d = CGL.newDataArray(
        g, CGK.CoordinateY_s, NPY.ones((5, 7, 9), dtype="float64", order="F")
    )
    d = CGL.newDataArray(
        g, CGK.CoordinateZ_s, NPY.ones((5, 7, 9), dtype="float64", order="F")
    )
    s = NPY.array([[vertexsize, cellsize, 0]], dtype="int32", order="F")
    z2 = CGL.newZone(b, "Zone2", s, CGK.Unstructured_s)
    g = CGL.newGridCoordinates(z2, "GridCoordinates")
    d = CGL.newDataArray(
        g, CGK.CoordinateX_s, NPY.ones((vertexsize), dtype="float64", order="F")
    )
    d = CGL.newDataArray(
        g, CGK.CoordinateY_s, NPY.ones((vertexsize), dtype="float64", order="F")
    )
    d = CGL.newDataArray(
        g, CGK.CoordinateZ_s, NPY.ones((vertexsize), dtype="float64", order="F")
    )
    tetras = CGL.newElements(
        z2,
        "TETRAS",
        CGK.TETRA_4_s,
        NPY.array([[1, cellsize]], "i", order="F"),
        NPY.ones((cellsize * 4), dtype="int32"),
    )
    tris = CGL.newElements(
        z2,
        "TRIS",
        CGK.TRI_3_s,
        NPY.array([[cellsize + 1, cellsize + ntris]], "i", order="F"),
        NPY.ones((ntris * 3), dtype="int32"),
    )
    z3 = CGU.copyNode(z1, "Zone3")
    b[2].append(z3)
    z4 = CGU.copyNode(z2, "Zone4")
    b[2].append(z4)
    z = [z1, z2, z3, z4]
    return (T, b, z)


(T, b, z) = makeCorrectTree()
zgc = CGL.newZoneGridConnectivity(z[0])
gc = CGL.newGridConnectivity(zgc, "join1_3", "Zone3", ctype=CGK.Abutting1to1_s)
CGL.newPointRange(gc, value=NPY.array([[1, 1], [1, 7], [1, 9]], order="F"))
CGL.newPointRange(
    gc, name=CGK.PointRangeDonor_s, value=NPY.array([[1, 1], [1, 7], [1, 9]], order="F")
)
gc = CGL.newGridConnectivity(zgc, "join1_2", "Zone2", ctype=CGK.Abutting1to1_s)
CGL.newPointRange(gc, value=NPY.array([[1, 5], [1, 1], [1, 9]], order="F"))
CGL.newIndexArray(
    gc,
    CGK.PointListDonor_s,
    value=NPY.array([range(cellsize + 1, cellsize + ntris + 1)], order="F"),
)
CGL.newGridLocation(gc, value=CGK.FaceCenter_s)
TESTS.append((tag, T, diag))

#  -------------------------------------------------------------------------
tag = "gridconnectivity bad datatype"
diag = False
(T, b, z) = makeCorrectTree()
zgc = CGL.newZoneGridConnectivity(z[0])
gc = CGL.newGridConnectivity(zgc, "join1_2", "Zone2", ctype=CGK.Abutting1to1_s)
CGL.newPointRange(gc, value=NPY.array([[1, 5], [1, 1], [1, 9]], order="F"))
CGL.newIndexArray(
    gc,
    CGK.PointListDonor_s,
    value=NPY.array([range(cellsize + 1, cellsize + ntris + 1)], order="F"),
)
CGL.newGridLocation(gc, value=CGK.FaceCenter_s)
gc[1] = NPY.array([1], order="F")
TESTS.append((tag, T, diag))

#  -------------------------------------------------------------------------
tag = "gridconnectivity unfound zonedonor"
diag = False
(T, b, z) = makeCorrectTree()
zgc = CGL.newZoneGridConnectivity(z[0])
gc = CGL.newGridConnectivity(zgc, "join1_2", "Zone6", ctype=CGK.Abutting1to1_s)
CGL.newPointRange(gc, value=NPY.array([[1, 5], [1, 1], [1, 9]], order="F"))
CGL.newIndexArray(
    gc,
    CGK.PointListDonor_s,
    value=NPY.array([range(cellsize + 1, cellsize + ntris + 1)], order="F"),
)
CGL.newGridLocation(gc, value=CGK.FaceCenter_s)
TESTS.append((tag, T, diag))

#  -------------------------------------------------------------------------
tag = "gridconnectivity bad pointrange/list"
diag = False
(T, b, z) = makeCorrectTree()
zgc = CGL.newZoneGridConnectivity(z[0])
gc = CGL.newGridConnectivity(zgc, "join1_2", "Zone2", ctype=CGK.Abutting1to1_s)
CGL.newPointRange(gc, value=NPY.array([[1, 5], [1, 1], [1, 9]], order="F"))
CGL.newIndexArray(
    gc, CGK.PointList_s, value=NPY.array([[1, 5], [1, 1], [1, 9]], order="F")
)
CGL.newIndexArray(
    gc,
    CGK.PointListDonor_s,
    value=NPY.array([range(cellsize + 1, cellsize + ntris + 1)], order="F"),
)
CGL.newGridLocation(gc, value=CGK.FaceCenter_s)
gc = CGL.newGridConnectivity(zgc, "join1_3", "Zone3", ctype=CGK.Abutting1to1_s)
CGL.newPointRange(
    gc, name=CGK.PointRangeDonor_s, value=NPY.array([[1, 1], [1, 7], [1, 9]], order="F")
)
TESTS.append((tag, T, diag))

#  -------------------------------------------------------------------------
tag = "gridconnectivity bad gridlocation"
diag = False
(T, b, z) = makeCorrectTree()
zgc = CGL.newZoneGridConnectivity(z[0])
gc = CGL.newGridConnectivity(zgc, "join1_2a", "Zone2", ctype=CGK.Abutting1to1_s)
CGL.newPointRange(gc, value=NPY.array([[1, 4], [1, 1], [1, 8]], order="F"))
CGL.newIndexArray(
    gc, CGK.PointListDonor_s, value=NPY.array([range(1, cellsize)], order="F")
)
CGL.newGridLocation(gc, value=CGK.CellCenter_s)
gc = CGL.newGridConnectivity(zgc, "join1_2b", "Zone2", ctype=CGK.Overset_s)
CGL.newPointRange(gc, value=NPY.array([[1, 5], [1, 7], [1, 1]], order="F"))
CGL.newIndexArray(
    gc,
    CGK.PointListDonor_s,
    value=NPY.array([range(cellsize + 1, cellsize + ntris + 1)], order="F"),
)
CGL.newGridLocation(gc, value=CGK.FaceCenter_s)
TESTS.append((tag, T, diag))

#  -------------------------------------------------------------------------
tag = "gridconnectivity bad range on zonedonor #1"
diag = False
(T, b, z) = makeCorrectTree()
zgc = CGL.newZoneGridConnectivity(z[0])
gc = CGL.newGridConnectivity(zgc, "join1_2", "Zone2", ctype=CGK.Abutting1to1_s)
CGL.newPointRange(gc, value=NPY.array([[1, 5], [1, 1], [1, 9]], order="F"))
CGL.newIndexArray(
    gc,
    CGK.PointListDonor_s,
    value=NPY.array([range(cellsize - 1, cellsize + ntris + 1)], order="F"),
)
CGL.newGridLocation(gc, value=CGK.FaceCenter_s)
TESTS.append((tag, T, diag))

#  -------------------------------------------------------------------------
tag = "gridconnectivity bad range on zonedonor #2"
diag = False
(T, b, z) = makeCorrectTree()
zgc = CGL.newZoneGridConnectivity(z[0])
gc = CGL.newGridConnectivity(zgc, "join1_2", "Zone2", ctype=CGK.Abutting1to1_s)
CGL.newPointRange(gc, value=NPY.array([[1, 5], [1, 1], [1, 9]], order="F"))
CGL.newIndexArray(
    gc,
    CGK.PointListDonor_s,
    value=NPY.array([range(cellsize + 2, cellsize + ntris + 2)], order="F"),
)
CGL.newGridLocation(gc, value=CGK.FaceCenter_s)
TESTS.append((tag, T, diag))

#  -------------------------------------------------------------------------
tag = "gridconnectivity bad range on zonedonor #3"
diag = False
(T, b, z) = makeCorrectTree()
zgc = CGL.newZoneGridConnectivity(z[0])
gc = CGL.newGridConnectivity(zgc, "join1_3", "Zone3", ctype=CGK.Overset_s)
CGL.newPointRange(gc, value=NPY.array([[1, 1], [1, 7], [1, 9]], order="F"))
CGL.newPointRange(
    gc, name=CGK.PointRangeDonor_s, value=NPY.array([[1, 1], [1, 7], [1, 9]], order="F")
)
CGL.newGridLocation(gc, value=CGK.CellCenter_s)
TESTS.append((tag, T, diag))

#  -------------------------------------------------------------------------
tag = "gridconnectivity abutting not a face"
diag = False
(T, b, z) = makeCorrectTree()
zgc = CGL.newZoneGridConnectivity(z[0])
gc = CGL.newGridConnectivity(zgc, "join1_3", "Zone3", ctype=CGK.Abutting1to1_s)
CGL.newPointRange(gc, value=NPY.array([[1, 5], [1, 7], [1, 9]], order="F"))
CGL.newPointRange(
    gc, name=CGK.PointRangeDonor_s, value=NPY.array([[1, 5], [1, 7], [1, 9]], order="F")
)
TESTS.append((tag, T, diag))



#!/usr/bin/env python

# [ name, value, children, type ]
import CGNS.MAP
import CGNS.PAT
import CGNS.PAT.cgnsutils

(tree, links) = CGNS.MAP.load(
    "./001Disk.hdf", CGNS.MAP.S2P_FOLLOWLINKS | CGNS.MAP.S2P_TRACE
)

ref = ["ReferenceState", None, [], CGNS.PAT.ReferenceState_t]
ref[2].append(["Density", dens, [], "DataArray_t"])

node = CGNS.PAT.cgnsutils.getNodeByPath(tree, "/Disk")
node[2].append(ref)

CGNS.MAP.save(
    "./001Disk-new.hdf", CGNS.MAP.S2P_FOLLOWLINKS | CGNS.MAP.S2P_TRACE, tree, links
)



#  -------------------------------------------------------------------------
#  pyCGNS - Python package for CFD General Notation System -
#  See license.txt file in the root directory of this Python module source
#  -------------------------------------------------------------------------
#

import sys
import time

from CGNS.pyCGNSconfig import HAS_MSW

if HAS_MSW:
    import ctypes

    myappid = "pycgns.pycgns.cglook.version"  # arbitrary string
    ctypes.windll.shell32.SetCurrentProcessExplicitAppUserModelID(myappid)

from CGNS.NAV.moption import Q7OptionContext as OCTXT

from qtpy import QtCore
from qtpy.QtWidgets import QApplication, QSplashScreen
from qtpy.QtGui import QPixmap

# from qtpy.QtGui import *
from CGNS.NAV.wcontrol import Q7Main

splash = None
wcontrol = None


def closeSplash():
    global splash
    global wcontrol
    splash.finish(wcontrol)


# -----------------------------------------------------------------
def run(
    args=[],
    files=[],
    datasets=[],
    flags=(False, False, False, False, False, False, False),
    ppath=None,
    query=None,
):
    hidecontrol = False
    global splash
    global wcontrol

    if flags[4]:
        from CGNS.NAV.wquery import Q7Query

        Q7Query.loadUserQueries()
        Q7Query.fillQueries()
        ql = Q7Query.queriesNamesList()
        print("# CGNS.NAV available queries:")
        for q in ql:
            print(" ", q)
    else:
        app = QApplication(args)
        if not flags[5]:
            pixmap = QPixmap(":/images/splash.png")
            splash = QSplashScreen(pixmap, QtCore.Qt.WindowStaysOnTopHint)
            splash.show()
            splash.showMessage(
                "Release v%s" % OCTXT._ToolVersion,
                QtCore.Qt.AlignHCenter | QtCore.Qt.AlignBottom,
            )
            timer = QtCore.QTimer.singleShot(3000, closeSplash)
        app.processEvents()
        Q7Main.verbose = flags[2]
        wcontrol = Q7Main()
        wcontrol._application = app
        wcontrol.setOptionValue("NAVTrace", flags[2])
        wcontrol.transientRecurse = flags[0]
        wcontrol.transientVTK = flags[3]
        wcontrol.query = query
        wcontrol._T("start")
        wcontrol.setDefaults()
        if flags[1]:
            wcontrol.loadlast()
            hidecontrol = flags[6]
        if files:
            for ff in files:
                wcontrol.loadfile(ff)
            hidecontrol = flags[6]
        if datasets:
            for dd in datasets:
                wcontrol.loadCompleted(
                    dataset_name="FFFF",
                    dataset_base="BASE",
                    dataset_tree=dd,
                    dataset_references=[],
                    dataset_paths=[],
                )
            hidecontrol = flags[6]
        wcontrol.show()
        if hidecontrol:
            wcontrol.hide()
        app.exec_()
        wcontrol._T("leave")
    sys.exit()


# run()

# -----------------------------------------------------------------



#  -------------------------------------------------------------------------
#  pyCGNS - Python package for CFD General Notation System
#  See license.txt file in the root directory of this Python module source  
#  -------------------------------------------------------------------------
# This file has been generated on [2022-04-08 12:20:09]
# Using platform [Windows fv-az200-1 Intel64 Family 6 Model 106 Stepping 6, GenuineIntel]

DATE="2022-04-08 12:20:09"
PLATFORM="Windows fv-az200-1 Intel64 Family 6 Model 106 Stepping 6, GenuineIntel"
PFX="#"

HAS_HDF5=True
HAS_H5PY=0
HAS_VTK=True
HAS_MSW=1
USE_COMPACT_STORAGE=1

INCLUDE_DIRS=['D:\\bld\\pycgns_1649419988797\\_h_env\\lib\\site-packages\\numpy\\core\\include', 'D:\\bld\\pycgns_1649419988797\\_h_env\\Library\\include', 'D:\\bld\\pycgns_1649419988797\\_h_env/lib/python3.9/site-packages/numpy/core/include', 'D:/bld/pycgns_1649419988797/_h_env/Library/include', '', 'D:\\bld\\pycgns_1649419988797\\_h_env/include', '../include', '/include']
LIBRARY_DIRS=['D:\\bld\\pycgns_1649419988797\\_h_env\\Library\\lib', 'D:/bld/pycgns_1649419988797/_h_env/Library/lib', '', 'D:\\bld\\pycgns_1649419988797\\_h_env/lib', '../lib', '/lib64', '/lib']

HDF5_VERSION="1.12.1"
HDF5_PATH_INCLUDES=['D:\\bld\\pycgns_1649419988797\\_h_env\\Library\\include']
HDF5_PATH_LIBRARIES=['D:\\bld\\pycgns_1649419988797\\_h_env\\Library\\lib']
HDF5_LINK_LIBRARIES=['hdf5']
HDF5_EXTRA_ARGS=[]

NUMPY_VERSION="1.19.5"
NUMPY_VERSION_API="0x01000009"
NUMPY_PATH_INCLUDES=['D:\\bld\\pycgns_1649419988797\\_h_env\\lib\\site-packages\\numpy\\core\\include']
NUMPY_PATH_LIBRARIES=['D:\\bld\\pycgns_1649419988797\\_h_env\\Library\\lib']
NUMPY_LINK_LIBRARIES=[]
NUMPY_EXTRA_ARGS=[]

PYQT_VERSION="2.0.1"
QT_VERSION="5.12.9"
CYTHON_VERSION="0.29.28"
VTK_VERSION="9.1.0"

__version__=6 # @@UPDATEVERSION@@
__release__=1 # @@UPDATERELEASE@@
__revision__=0 # @@UPDATEREVISION@@
__vid__="%s.%s.%s"%(__version__,__release__,__revision__)
__doc__="pyCGNS - %s - Python package for CGNS (CFD General Notation System)"        %(__vid__)
version=__vid__

NAME="pyCGNS"
VERSION=__vid__
DESCRIPTION=__doc__
AUTHOR="Marc Poinot"
EMAIL="marc.poinot@safrangroup.com"
LICENSE="LGPL 2"




#  -------------------------------------------------------------------------
#  pyCGNS - Python package for CFD General Notation System -
#  See license.txt file in the root directory of this Python module source
#  -------------------------------------------------------------------------
#
"""
 We are changing for a full h5py implementation, default switch is in
 config: HAS_H5PY forces h5py instead of raw HDF5 C API implementation

 CGNS.MAP module is the interface to CGNS/HDF5 (formerly was CHLone)
  - now all the CHLone code is embedded in CGNS.MAP (EmbeddedCHLone)
  - you DO NOT need to install CHLone on its own

  Flags: 

  - python does NOT have bitfiled or unseigned int that would act as
    bit masks we have in C. So each operation on such bitfields (flags)
    should be re-masked against the actual 32bits size of the flags,
    which means you should always have a & flags.ALL in your operation
"""
#
from CGNS.pyCGNSconfig import HAS_H5PY

if HAS_H5PY:
    from .cgio import load, save, probe, flags

    NONE = flags.NONE
    ALL = flags.ALL
    TRACE = flags.TRACE
    FOLLOWLINKS = flags.FOLLOWLINKS
    DEBUG = flags.DEBUG
    NODATA = flags.NODATA
    MERGELINKS = 0
    COMPRESS = flags.COMPRESS
    REVERSEDIMS = flags.REVERSEDIMS
    OWNDATA = flags.OWNDATA
    UPDATE = flags.UPDATE
    DELETEMISSING = flags.DELETEMISSING
    ALTERNATESIDS = flags.ALTERNATESIDS
    NOTRANSPOSE = 0
    UPDATEONLY = flags.UPDATEONLY
    FORTRANFLAG = flags.FORTRANFLAG
    PROPAGATE = flags.PROPAGATE
    DEFAULT = flags.DEFAULT
    DEFAULTS = flags.DEFAULTS
    KEEPLIST = flags.KEEPLIST
    CHECKSUM = flags.CHECKSUM

    LKOK = flags.links.OK
    LKFAIL = flags.links.FAIL
    LKBADSYNTAX = flags.links.BADSYNTAX
    LKNOFILE = flags.links.NOFILE
    LKFILENOREAD = flags.links.FILENOREAD
    LKNONODE = flags.links.NONODE
    LKLOOP = flags.links.LOOP
    LKIGNORED = flags.links.IGNORED

else:
    from .EmbeddedCHLone import load, save, probe
    from .EmbeddedCHLone import CHLoneException as error

    #
    from . import EmbeddedCHLone as CHL

    #
    S2P_NONE = CHL.FNONE
    S2P_ALL = CHL.FALL
    S2P_TRACE = CHL.FTRACE
    S2P_FOLLOWLINKS = CHL.FFOLLOWLINKS
    S2P_DEBUG = CHL.FDEBUG
    S2P_NODATA = CHL.FNODATA
    S2P_MERGELINKS = 0
    S2P_COMPRESS = CHL.FCOMPRESS
    S2P_REVERSEDIMS = CHL.FREVERSEDIMS
    S2P_OWNDATA = CHL.FOWNDATA
    S2P_UPDATE = CHL.FUPDATE
    S2P_DELETEMISSING = CHL.FDELETEMISSING
    S2P_ALTERNATESIDS = CHL.FALTERNATESIDS
    S2P_NOTRANSPOSE = 0
    S2P_UPDATEONLY = CHL.FUPDATEONLY
    S2P_FORTRANFLAG = CHL.FFORTRANFLAG
    S2P_PROPAGATE = CHL.FPROPAGATE
    S2P_DEFAULT = CHL.FDEFAULT
    S2P_DEFAULTS = CHL.FDEFAULTS
    S2P_KEEPLIST = CHL.FKEEPLIST
    S2P_CHECKSUM = CHL.FCHECKSUM

    S2P_LKOK = CHL.LKOK
    S2P_LKFAIL = CHL.LKFAIL
    S2P_LKBADSYNTAX = CHL.LKBADSYNTAX
    S2P_LKNOFILE = CHL.LKNOFILE
    S2P_LKFILENOREAD = CHL.LKFILENOREAD
    S2P_LKNONODE = CHL.LKNONODE
    S2P_LKLOOP = CHL.LKLOOP
    S2P_LKIGNORED = CHL.LKIGNORED

    NONE = CHL.FNONE
    ALL = CHL.FALL
    TRACE = CHL.FTRACE
    FOLLOWLINKS = CHL.FFOLLOWLINKS
    DEBUG = CHL.FDEBUG
    NODATA = CHL.FNODATA
    MERGELINKS = 0
    COMPRESS = CHL.FCOMPRESS
    REVERSEDIMS = CHL.FREVERSEDIMS
    OWNDATA = CHL.FOWNDATA
    UPDATE = CHL.FUPDATE
    DELETEMISSING = CHL.FDELETEMISSING
    ALTERNATESIDS = CHL.FALTERNATESIDS
    NOTRANSPOSE = 0
    UPDATEONLY = CHL.FUPDATEONLY
    FORTRANFLAG = CHL.FFORTRANFLAG
    PROPAGATE = CHL.FPROPAGATE
    DEFAULT = CHL.FDEFAULT
    DEFAULTS = CHL.FDEFAULTS
    KEEPLIST = CHL.FKEEPLIST
    CHECKSUM = CHL.FCHECKSUM

    LKOK = CHL.LKOK
    LKFAIL = CHL.LKFAIL
    LKBADSYNTAX = CHL.LKBADSYNTAX
    LKNOFILE = CHL.LKNOFILE
    LKFILENOREAD = CHL.LKFILENOREAD
    LKNONODE = CHL.LKNONODE
    LKLOOP = CHL.LKLOOP
    LKIGNORED = CHL.LKIGNORED

    # --- forward compat stuff here
    class Flags(object):
        pass

    flags = Flags()
    flags.links = Flags()

    flags.NONE = NONE
    flags.ALL = ALL
    flags.TRACE = TRACE
    flags.FOLLOWLINKS = FOLLOWLINKS
    flags.DEBUG = DEBUG
    flags.NODATA = NODATA
    flags.MERGELINKS = 0
    flags.COMPRESS = COMPRESS
    flags.REVERSEDIMS = REVERSEDIMS
    flags.OWNDATA = OWNDATA
    flags.UPDATE = UPDATE
    flags.DELETEMISSING = DELETEMISSING
    flags.ALTERNATESIDS = ALTERNATESIDS
    flags.NOTRANSPOSE = 0
    flags.UPDATEONLY = UPDATEONLY
    flags.FORTRANFLAG = FORTRANFLAG
    flags.PROPAGATE = PROPAGATE
    flags.DEFAULT = DEFAULT
    flags.DEFAULTS = DEFAULTS
    flags.KEEPLIST = KEEPLIST
    flags.CHECKSUM = CHECKSUM

    flags.links.OK = LKOK
    flags.links.FAIL = LKFAIL
    flags.links.BADSYNTAX = LKBADSYNTAX
    flags.links.NOFILE = LKNOFILE
    flags.links.FILENOREAD = LKFILENOREAD
    flags.links.NONODE = LKNONODE
    flags.links.LOOP = LKLOOP
    flags.links.IGNORED = LKIGNORED


def flags_set(source=flags.DEFAULT, flag=flags.NONE):
    return source | flag & flags.ALL


def flags_unset(source, flag):
    return source & ~flag & flags.ALL


def flags_check(source, flag):
    return source & flag == flag


# --- last line



#  -------------------------------------------------------------------------
#  pyCGNS.VAL - Python package for CFD General Notation System - VALidater
#  See license.txt file in the root directory of this Python module source
#  -------------------------------------------------------------------------
#
import CGNS.PAT.cgnslib as CGL
import CGNS.PAT.cgnsutils as CGU
import CGNS.PAT.cgnskeywords as CGK
import numpy as NPY

TESTS = []

#  -------------------------------------------------------------------------
tag = "bc"
diag = True


def makeCorrectTree(vertexsize, cellsize, ntris):
    T = CGL.newCGNSTree()
    b = CGL.newBase(T, "Base", 3, 3)
    s = NPY.array([[vertexsize, cellsize, 0]], dtype="int32", order="F")
    z = CGL.newZone(b, "Zone", s, CGK.Unstructured_s)
    g = CGL.newGridCoordinates(z, "GridCoordinates")
    d = CGL.newDataArray(
        g, CGK.CoordinateX_s, NPY.ones((vertexsize), dtype="float64", order="F")
    )
    d = CGL.newDataArray(
        g, CGK.CoordinateY_s, NPY.ones((vertexsize), dtype="float64", order="F")
    )
    d = CGL.newDataArray(
        g, CGK.CoordinateZ_s, NPY.ones((vertexsize), dtype="float64", order="F")
    )
    tetras = CGL.newElements(
        z,
        "TETRAS",
        CGK.TETRA_4_s,
        NPY.array([[1, cellsize]], "i", order="F"),
        NPY.ones((cellsize * 4), dtype="int32"),
    )
    tris = CGL.newElements(
        z,
        "TRIS",
        CGK.TRI_3_s,
        NPY.array([[cellsize + 1, cellsize + ntris]], "i", order="F"),
        NPY.ones((ntris * 3), dtype="int32"),
    )
    zbc = CGL.newZoneBC(z)
    n = CGL.newBoundary(
        zbc,
        "BC",
        [range(cellsize + 1, cellsize + ntris + 1)],
        btype=CGK.Null_s,
        family=None,
        pttype=CGK.PointList_s,
    )
    g = CGL.newGridLocation(n, value=CGK.FaceCenter_s)
    return (T, b, z, zbc, n, g)


vertexsize = 20
cellsize = 7
ntris = 12
(T, b, z, zbc, n, g) = makeCorrectTree(vertexsize, cellsize, ntris)
TESTS.append((tag, T, diag))

#  -------------------------------------------------------------------------
tag = "bc bad location"
diag = False
(T, b, z, zbc, n, g) = makeCorrectTree(vertexsize, cellsize, ntris)
CGL.newPointRange(n, value=NPY.array([[1, cellsize]], "i"))
zbc[2] = []
n = CGL.newBoundary(
    zbc,
    "BC",
    [range(1, cellsize + 1)],
    btype=CGK.Null_s,
    family=None,
    pttype=CGK.PointList_s,
)
g = CGL.newGridLocation(n, value=CGK.CellCenter_s)
TESTS.append((tag, T, diag))

#  -------------------------------------------------------------------------
tag = "bc both PointList and PointRange"
diag = False
(T, b, z, zbc, n, g) = makeCorrectTree(vertexsize, cellsize, ntris)
CGL.newPointRange(n, value=NPY.array([[cellsize + 1, cellsize + ntris]], "i"))
TESTS.append((tag, T, diag))

#  -------------------------------------------------------------------------
tag = "bc no PointList or PointRange"
diag = False
(T, b, z, zbc, n, g) = makeCorrectTree(vertexsize, cellsize, ntris)
CGU.removeChildByName(n, CGK.PointList_s)
TESTS.append((tag, T, diag))

#  -------------------------------------------------------------------------
tag = "bc FamilySpecified but not FamilyName"
diag = False
(T, b, z, zbc, n, g) = makeCorrectTree(vertexsize, cellsize, ntris)
n[1] = CGU.setStringAsArray(CGK.FamilySpecified_s)
TESTS.append((tag, T, diag))



#  -------------------------------------------------------------------------
#  pyCGNS - Python package for CFD General Notation System -
#  See license.txt file in the root directory of this Python module source
#  -------------------------------------------------------------------------
#
from CGNS.NAV.moption import Q7OptionContext as OCTXT

import CGNS.PAT.cgnsutils as CGU
import CGNS.PAT.cgnskeywords as CGK

from qtpy.QtCore import Qt, QModelIndex
from qtpy.QtWidgets import QStyledItemDelegate

from CGNS.NAV.Q7FormWindow import Ui_Q7FormWindow
from CGNS.NAV.mtable import Q7TableModel
from CGNS.NAV.wstylesheets import Q7TABLEVIEWSTYLESHEET
from CGNS.NAV.wfingerprint import Q7Window


# -----------------------------------------------------------------
def divpairs(n):
    d = n
    l = []
    while d != 0:
        m = n // d
        r = n % d
        if r == 0:
            l += [(d, m)]
        d -= 1
    return l


# -----------------------------------------------------------------
class Q7TableItemDelegate(QStyledItemDelegate):
    def paint(self, painter, option, index):
        t = index.data(Qt.DisplayRole)
        r = option.rect.adjusted(2, 2, -2, -2)
        painter.drawText(r, Qt.AlignVCenter | Qt.AlignLeft, t, r)
        # QStyledItemDelegate.paint(self, painter, option, index)


# -----------------------------------------------------------------
class Q7Form(Q7Window, Ui_Q7FormWindow):
    def __init__(self, control, node, fgprintindex):
        Q7Window.__init__(
            self, Q7Window.VIEW_FORM, control, node.sidsPath(), fgprintindex
        )
        self._node = node
        self._operatorlist = [""]
        for t in self._node.sidsTypeList():
            self.eType.addItem(t)
        self.bClose.clicked.connect(self.reject)
        self.bInfo.clicked.connect(self.infoFormView)
        self.cFortranOrderOff.setChecked(False)
        self.cFortranOrderOff.setDisabled(True)
        for dt in node.sidsDataType(all=True):
            self.cDataType.addItem(dt)

    def infoFormView(self):
        self._control.helpWindow("Form")

    def updateTreeStatus(self):
        pass

    def accept(self):
        pass

    def reject(self):
        self.close()

    def show(self):
        self.reset()
        super(Q7Form, self).show()

    def reset(self):
        self.eName.setText(self._node.sidsName())
        self.ePath.setText(self._node.sidsPath())
        self.ePath.setReadOnly(True)
        self.sMinH.setRange(-999999999, 999999999)
        self.sMinV.setRange(-999999999, 999999999)
        self.sMinH.setValue(1)
        self.sMinV.setValue(1)
        self.setCurrentType(self._node.sidsType())
        dims = self._node.sidsDims()
        its = 1
        for x in dims:
            its *= x
        ndz = its * self._node.sidsDataTypeSize()
        self.eDims.setText(str(dims))
        self.setDataType(self._node)
        self.eItems.setText(str(its))
        self.ls = 1
        self.cs = dims[0]
        if len(dims) > 1:
            self.ls = 1
            for x in dims[1:]:
                self.ls *= x
        self.showparams = {"cs": self.cs, "ls": self.ls, "mode": "IJ"}
        for dp in divpairs(self.cs * self.ls):
            self.cRowColSize.addItem("%d x %d" % dp)
        ix = self.cRowColSize.findText("%d x %d" % (self.cs, self.ls))
        self.cRowColSize.setCurrentIndex(ix)
        self.model = Q7TableModel(self._node, self.showparams)
        self.tableView.setModel(self.model)
        self.tableView.setStyleSheet(self._stylesheet)
        self.bMinimize.clicked.connect(self.minimizeCells)
        #
        self.tableView.clicked[QModelIndex].connect(self.clickedNode)
        self.cRowColSize.currentIndexChanged[int].connect(self.resizeTable)
        self.sMinH.valueChanged[int].connect(self.resetIndex)
        self.sMinV.valueChanged[int].connect(self.resetIndex)
        # QObject.connect(self.tableView,
        #                SIGNAL("clicked(QModelIndex)"),
        #                self.clickedNode)
        # QObject.connect(self.cRowColSize,
        #                SIGNAL("currentIndexChanged(int)"),
        #                self.resizeTable)
        # QObject.connect(self.sMinH,
        #                SIGNAL("valueChanged(int)"),
        #                self.resetIndex)
        # QObject.connect(self.sMinV,
        #                SIGNAL("valueChanged(int)"),
        #                self.resetIndex)
        lk = self.FG.isLink(self._node.sidsPath())
        if lk:
            self.setLabel(self.eDestDir, lk[0])
            self.setLabel(self.eDestFile, lk[1])
            self.setLabel(self.eDestPath, lk[2])
        else:
            self.tFiles.setDisabled(True)
        self.tPython.setDisabled(True)
        if self._node.sidsDataType() == CGK.C1:
            if self._node.sidsValue().ndim > 1:
                txt = self._node.sidsValue().T.tostring().decode("ascii")
                self.cFortranOrderOff.setChecked(True)
                self.cFortranOrderOff.setDisabled(True)
            else:
                txt = self._node.sidsValue().tostring().decode("ascii")
            self.eText.initText(txt)

    def resizeTable(self):
        s = self.cRowColSize.currentText()
        (r, c) = s.split("x")
        self.cs = int(r)
        self.ls = int(c)
        self.showparams = {"cs": self.cs, "ls": self.ls, "mode": "IJ"}
        self.model = Q7TableModel(self._node, self.showparams)
        self.tableView.setModel(self.model)
        self.tableView.setStyleSheet(self._stylesheet)

    def clickedNode(self, index):
        pass

    def minimizeCells(self, *args):
        self.tableView.resizeColumnsToContents()
        self.tableView.resizeRowsToContents()

    def setCurrentType(self, ntype):
        idx = self.eType.findText(ntype)
        self.eType.setCurrentIndex(idx)

    def setDataType(self, node):
        dt = node.sidsDataType()
        ix = self.cDataType.findText(dt)
        if ix == -1:
            return
        self.cDataType.setCurrentIndex(ix)

    def resetIndex(self):
        hm = self.sMinH.value()
        vm = self.sMinV.value()
        self.model.setRange(hm, vm)
        self.model.headerDataChanged.emit(Qt.Vertical, 1, 1)
        self.model.headerDataChanged.emit(Qt.Horizontal, 1, 1)

    def addControlLine(self):
        pass

    def doRelease(self):
        pass


# -----------------------------------------------------------------



#  ---------------------------------------------------------------------------
#  pyCGNS - Python package for CFD General Notation System -
#  See license.txt file in the root directory of this Python module source
#  ---------------------------------------------------------------------------
#
import CGNS.PAT.cgnslib as C
import CGNS.PAT.cgnserrors as E
import CGNS.PAT.cgnskeywords as K
import numpy as N

#
data = C.newFamily(None, "{Family}")
C.newGeometryReference(data, "{GeometryReference}")
C.newRotatingCoordinates(data)
C.newUserDefinedData(data, "{UserDefinedData}")
C.newDescriptor(data, "{Descriptor}")
C.newOrdinal(data)

#
status = "12.6"
comment = "Full SIDS with all optionals"
pattern = [data, status, comment]
#



#  ---------------------------------------------------------------------------
#  pyCGNS - Python package for CFD General Notation System -
#  See license.txt file in the root directory of this Python module source
#  ---------------------------------------------------------------------------
#
import CGNS.PAT.cgnslib as C
import CGNS.PAT.cgnserrors as E
import CGNS.PAT.cgnskeywords as K
import numpy as N

data = C.newCoordinates(None)
C.newRind(data, N.array([0, 0, 0, 0, 1, 1]))
C.newDataClass(data)
C.newDimensionalUnits(data)
C.newUserDefinedData(data, "{UserDefinedData}")
C.newDescriptor(data, "{Descriptor}")

C.newDataArray(data, K.CoordinateX_s)
C.newDataArray(data, K.CoordinateY_s)
C.newDataArray(data, K.CoordinateZ_s)

status = "7.1"
comment = "Full SIDS with all optionals"
pattern = [data, status, comment]



#  -------------------------------------------------------------------------
#  pyCGNS - Python package for CFD General Notation System -
#  See license.txt file in the root directory of this Python module source
#  -------------------------------------------------------------------------
#
import sys
import os
import os.path
import fnmatch
import imp
import CGNS.VAL.grammars.etablesids as STB

PROFILENAME = "grammars"


#  -------------------------------------------------------------------------
def readProfile():
    try:
        hdir = os.environ["HOME"]
    except:
        return {}
    pdir = "%s%s.CGNS.NAV" % (hdir, os.path.sep)
    if not os.path.exists(pdir):
        return {}
    sys.path.append(pdir)
    fp, pth, des = imp.find_module("grammars")
    try:
        mod = imp.load_module("grammars", fp, pth, des)
    finally:
        if fp:
            fp.close()
    return mod.Grammars


#  -------------------------------------------------------------------------
def findAllUserGrammars(verbose=False):
    kdict = {}
    for pth in [p for p in sys.path if p != ""]:
        if verbose:
            print("### scanning", pth)
        try:
            for pthroot, dirs, files in os.walk(pth):
                for fn in files:
                    if fnmatch.fnmatch(fn, "CGNS_VAL_USER_*.py") or fnmatch.fnmatch(
                        fn, "CGNS_VAL_USER_*.so"
                    ):
                        gkey = fn[14:-3]
                        if gkey in kdict:
                            if verbose:
                                print(
                                    "### * found grammar:",
                                )
                                print(gkey, "already found, ignore this one")
                        else:
                            if verbose:
                                print("### * found grammar:", gkey)
                            kdict[fn[14:-3]] = pthroot
                        if verbose:
                            print("### * found in :", pthroot)
                            if pthroot not in sys.path:
                                print("### * previous path is NOT in PYTHONPATH")
                            else:
                                print("### * previous path already is in PYTHONPATH")

        except OSError:
            pass
    return kdict


#  -------------------------------------------------------------------------
def findOneUserGrammar(tag, verbose=False):
    kdict = {}
    found = False
    for pth in sys.path:
        if verbose:
            print("### scanning", pth)
        try:
            for pthroot, dirs, files in os.walk(pth):
                for fn in files:
                    if fnmatch.fnmatch(fn, "CGNS_VAL_USER_%s.py" % tag):
                        kdict[fn[14:-3]] = pthroot
                        found = True
                        break
                if found:
                    break
        except OSError:
            pass
        if found:
            break
    return kdict


#  -------------------------------------------------------------------------
def importUserGrammars(key, recurse=False, verbose=False):
    mod = None
    modname = "CGNS_VAL_USER_%s" % key
    ipath = os.path.split(STB.__file__)[0]
    sys.path.append(ipath)

    if verbose:
        print("### Looking for grammar [%s]" % key)
    try:
        tp = imp.find_module(modname)
    except ImportError:
        if verbose:
            print("### Error: grammar [%s] not found" % key)
        if recurse:
            dk = findOneUserGrammar(key)
            if key in dk:
                sys.path.append(dk[key])
                if verbose:
                    print("### Warning: not in search path [%s]" % dk[key])
                try:
                    tp = imp.find_module(modname)
                except ImportError:
                    return None
            else:
                return None
        else:
            return None
    try:
        fp = tp[0]
        if tp[2][2] != imp.C_EXTENSION:
            mod = imp.load_module(modname, *tp)
        else:
            # print '### CGNS.VAL [info]: Module info',tp
            mod = imp.load_dynamic(modname, tp[1], tp[0])
    except:
        pass
    finally:
        if fp:
            fp.close()
    return mod


#  -------------------------------------------------------------------------
def locateGrammars():
    glist = []
    for k in ["SIDS", "elsA"]:
        mod = importUserGrammars(k)
        if mod is not None:
            glist.append((k, mod.__file__))
    return glist


# ---



import CGNS.PAT.cgnsutils as CGU
import CGNS.PAT.cgnskeywords as CGK


def getBCbyFamily(T, family):
    """Finds all BC that belongs to a family.
    Works with more than one base.
    Returns a list of path, you have to use getNodeByPath to retrieve
    the actual node."""
    result = []
    fpath = [
        CGK.CGNSTree_ts,
        CGK.CGNSBase_t,
        CGK.Zone_t,
        CGK.ZoneBC_t,
        CGK.BC_t,
        CGK.FamilyName_t,
    ]
    flist = CGU.getAllNodesByTypeOrNameList(T, fpath)
    for famnamepath in flist:
        fnode = CGU.getNodeByPath(T, famnamepath)
        if CGU.stringValueMatches(fnode, family):
            bcpath = CGU.getAncestor(famnamepath)
            result.append(bcpath)
    return result



#  -------------------------------------------------------------------------
#  pyCGNS - Python package for CFD General Notation System -
#  See license.txt file in the root directory of this Python module source
#  -------------------------------------------------------------------------
#
import numpy
import h5py
import os

from ..PAT import cgnskeywords as CGK
from ..PAT import cgnsutils as CGU
from ..PAT import cgnserrors as CGE


class Flags(object):
    pass


flags = Flags()
flags.NONE = 0x00000000
flags.ALL = 0xFFFFFFFF
flags.TRACE = 0x00000001
flags.FOLLOWLINKS = 0x00000002
flags.NODATA = 0x00000004
flags.KEEPLIST = 0x00000008
flags.MERGELINKS = 0
flags.COMPRESS = 0x00000010
flags.REVERSEDIMS = 0x00000020
flags.OWNDATA = 0x00000040
flags.UPDATE = 0x00000080
flags.DEBUG = 0x00008000
flags.DELETEMISSING = 0x00000100
flags.ALTERNATESIDS = 0x00000200
flags.NOTRANSPOSE = 0
flags.UPDATEONLY = 0x00000400
flags.FORTRANFLAG = 0x00000800
flags.PROPAGATE = 0x00004000
flags.LINKOVERRIDE = 0x00010000
flags.DEFAULT = (
    flags.FOLLOWLINKS
    | flags.DELETEMISSING
    | flags.OWNDATA
    | flags.REVERSEDIMS
    | flags.FORTRANFLAG
    | flags.OWNDATA
    | flags.LINKOVERRIDE
)
flags.DEFAULTS = flags.DEFAULT
flags.CHECKSUM = 0x00080000

flags.links = Flags()
flags.links.OK = 0x0000
flags.links.FAIL = 0x0001
flags.links.BADSYNTAX = 0x0002
flags.links.NOFILE = 0x0004
flags.links.FILENOREAD = 0x0008
flags.links.NONODE = 0x0010
flags.links.LOOP = 0x0020
flags.links.IGNORED = 0x0040
flags.links.UPDATED = 0x0080


def raiseException(code, *args):
    if args:
        raise MAPException((code, MAPException.mTable[code].format(args)))
    else:
        raise MAPException((code, MAPException.mTable[code]))


def probe(filename, path=None):
    tfile = os.path.normpath(os.path.expanduser(filename))
    if not os.path.isfile(tfile):
        raiseException(900, tfile)
    return True


def load(filename, nodata=False, maxdata=65, subtree="", follow_links=False, **kwargs):
    """
    Load a CGNS/HDF5 CGNS tree as a CGNS/Python object::

      from CGNS.MAP import load

      (tree, links, paths) = load('cgnsfile.cgns')

    Detailed documentation about the load can be found
    `here <https://pycgns.github.io/MAP/_index.html>`_, the following doc only
    describes the function signature.

    :arg str filename: CGNS/HDF5 file to open
    :arg bool nodata: if `True` the actual node data is load only if below the `maxdata`
    :arg int maxdata: if `nodata` is `True`, `maxdata` gives the threshold for loading actual data
    :arg str subtree: loads only starting from the `subtree` path arg
    :arg bool follow_links: If `True` follows the linked-to files (default:`False`)

    :return: a tuple (tree, links, paths) see remaks below

    :Remarks:
      - `tree` is a Python list, the actual CGNS/Python tree created from the CGNS/HDF5 file.
      - `links` is the list of linked-to files and nodes used during the `tree` creation.
      - `paths` is the list of nodes paths in `tree` you have to take care of such as no-data nodes.

    """

    def _load(grp, links, paths, load_children=True):

        if grp.attrs["type"] == "LK":
            if not follow_links:
                links.append(
                    [
                        "",
                        grp[" file"].value[:-1].tostring().decode("ascii"),
                        grp[" path"].value[:-1].tostring().decode("ascii"),
                        grp.name,
                        64,
                    ]
                )
                return None
            else:
                grp = grp[" link"]

        node = [unicode(grp.attrs["name"]), None, [], unicode(grp.attrs["label"])]

        if " data" in grp:
            dset = grp[" data"]
            n = dset.size
            if not nodata or n < maxdata:
                load_data = True
            else:
                load_data = False
                paths.append([grp.name, 1, grp.attrs["type"], dset.shape])

            if load_data:
                node[1] = numpy.copy(dset.value.transpose(), order="F")
                if grp.attrs["type"] == "C1":
                    node[1] = numpy.vectorize(chr)(node[1]).astype(numpy.dtype("c"))

        if load_children:
            cnames = [x for x in grp.keys() if x[0] != " "]
            for cname in cnames:
                child = _load(grp[cname], links, paths)
                if child is not None:
                    node[2].append(child)

        return node

    links = []
    paths = []
    with h5py.File(filename, "r") as h5f:
        if subtree == "":
            tree = _load(h5f, links, paths)
        else:
            grp = h5f[subtree]
            node = _load(grp, links, paths)
            while grp.name != "/":
                pnode = _load(grp.parent, links, paths, load_children=False)
                pnode[2].append(node)
                node = pnode
                grp = grp.parent

            tree = node

    tree[0] = "CGNSTree"
    tree[3] = "CGNSTree_t"

    return tree, links, paths


def save(filename, tree, links=[], **kwargs):
    def _cst_size_str(s, n):
        return numpy.bytes_(bytes(s, "ascii").ljust(n, b"\x00"))

    def _save(grp, node):

        _cst_size_str(node[0], 33)

        if node[3] == "CGNSTree_t":
            grp.attrs["name"] = _cst_size_str("HDF5 MotherNode", 33)
            grp.attrs["label"] = _cst_size_str("Root Node of HDF5 File", 33)
        else:
            grp.attrs["name"] = _cst_size_str(node[0], 33)
            grp.attrs["label"] = _cst_size_str(node[3], 33)
            grp.attrs["flags"] = numpy.array([0], dtype=numpy.int32)

        if node[1] is None:
            grp.attrs["type"] = _cst_size_str("MT", 3)
        else:
            if node[1].dtype == numpy.float32:
                grp.attrs["type"] = _cst_size_str("R4", 3)
            elif node[1].dtype == numpy.float64:
                grp.attrs["type"] = _cst_size_str("R8", 3)
            elif node[1].dtype == numpy.int32:
                grp.attrs["type"] = _cst_size_str("I4", 3)
            elif node[1].dtype == numpy.int64:
                grp.attrs["type"] = _cst_size_str("I8", 3)
            elif node[1].dtype == numpy.dtype("|S1"):
                grp.attrs["type"] = _cst_size_str("C1", 3)
            else:
                raise NotImplementedError("unknown dtype %s" % node[1].dtype)

            data = node[1]
            if node[1].dtype == numpy.dtype("|S1"):
                idx1 = numpy.nonzero(data != "")
                idx2 = numpy.nonzero(data == "")
                tmp = data.copy()
                data = numpy.zeros(data.shape, dtype=numpy.int8)
                data[idx1] = numpy.vectorize(ord)(tmp[idx1])
                data[idx2] = 0
                data = data.astype(numpy.int8)

            data = data.transpose()
            grp.create_dataset(" data", data=data)

        for child in node[2]:
            sgrp = grp.create_group(child[0])
            _save(sgrp, child)

    def _add_links(h5f, links):

        for dname, filename, rpth, lpth, n in links:
            tmp = lpth.split("/")
            name = tmp[-1]
            ppth = "/".join(tmp[:-1])

            grp = h5f[ppth]
            sgrp = grp.create_group(name)

            sgrp.attrs["name"] = _cst_size_str(name, 33)
            sgrp.attrs["label"] = _cst_size_str("", 33)
            sgrp.attrs["type"] = _cst_size_str("LK", 3)

            data = numpy.array([ord(x) for x in rpth + "\x00"], dtype=numpy.int8)
            sgrp.create_dataset(" path", data=data)
            data = numpy.array([ord(x) for x in filename + "\x00"], dtype=numpy.int8)
            sgrp.create_dataset(" file", data=data)
            sgrp[" link"] = h5py.ExternalLink(filename, rpth)

            assert dname == ""

    with h5py.File(filename, "w") as h5f:
        h5f.create_dataset(
            " format", data=numpy.array([78, 65, 84, 73, 86, 69, 0], dtype=numpy.int8)
        )
        h5f.create_dataset(
            " hdf5version",
            data=numpy.array(
                [
                    72,
                    68,
                    70,
                    53,
                    32,
                    86,
                    101,
                    114,
                    115,
                    105,
                    111,
                    110,
                    32,
                    49,
                    46,
                    56,
                    46,
                    49,
                    53,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                    0,
                ],
                dtype=numpy.int8,
            ),
        )
        _save(h5f, tree)
        _add_links(h5f, links)


#
# --- last line



#  -------------------------------------------------------------------------
#  pyCGNS.VAL - Python package for CFD General Notation System - VALidater
#  See license.txt file in the root directory of this Python module source
#  -------------------------------------------------------------------------
#
import CGNS.PAT.cgnslib as CGL
import CGNS.PAT.cgnsutils as CGU
import CGNS.PAT.cgnskeywords as CGK
import numpy as NPY

TESTS = []

#  -------------------------------------------------------------------------
tag = "zonesubregion"
diag = True


def makeCorrectTree(vertexsize, cellsize, ntris):
    T = CGL.newCGNSTree()
    b = CGL.newBase(T, "Base", 3, 3)
    s = NPY.array([[vertexsize, cellsize, 0]], dtype="int32", order="F")
    z = CGL.newZone(b, "Zone", s, CGK.Unstructured_s)
    g = CGL.newGridCoordinates(z, "GridCoordinates")
    d = CGL.newDataArray(
        g, CGK.CoordinateX_s, NPY.ones((vertexsize), dtype="float64", order="F")
    )
    d = CGL.newDataArray(
        g, CGK.CoordinateY_s, NPY.ones((vertexsize), dtype="float64", order="F")
    )
    d = CGL.newDataArray(
        g, CGK.CoordinateZ_s, NPY.ones((vertexsize), dtype="float64", order="F")
    )
    tetras = CGL.newElements(
        z,
        "TETRAS",
        CGK.TETRA_4_s,
        NPY.array([[1, cellsize]], "i", order="F"),
        NPY.ones((cellsize * 4), dtype="int32"),
    )
    tris = CGL.newElements(
        z,
        "TRIS",
        CGK.TRI_3_s,
        NPY.array([[cellsize + 1, cellsize + ntris]], "i", order="F"),
        NPY.ones((ntris * 3), dtype="int32"),
    )
    zbc = CGL.newZoneBC(z)
    n = CGL.newBoundary(
        zbc,
        "BC",
        [range(cellsize + 1, cellsize + ntris + 1)],
        btype=CGK.Null_s,
        family=None,
        pttype=CGK.PointList_s,
    )
    g = CGL.newGridLocation(n, value=CGK.FaceCenter_s)
    return (T, b, z)


vertexsize = 20
cellsize = 7
ntris = 11
(T, b, z) = makeCorrectTree(vertexsize, cellsize, ntris)
zsr1 = CGU.newNode(
    "SubRegion1",
    NPY.array([3], dtype=NPY.int32, order="F"),
    [],
    CGK.ZoneSubRegion_ts,
    parent=z,
)
CGU.newNode(
    CGK.PointRange_s,
    NPY.array([[1, vertexsize]], dtype=NPY.int32, order="F"),
    [],
    CGK.IndexRange_ts,
    parent=zsr1,
)
CGL.newDataArray(zsr1, "var", value=NPY.ones(vertexsize, dtype="float64", order="F"))
zsr2 = CGU.newNode(
    "SubRegion2",
    NPY.array([3], dtype=NPY.int32, order="F"),
    [],
    CGK.ZoneSubRegion_ts,
    parent=z,
)
CGU.newNode(
    CGK.PointList_s,
    NPY.array([range(1, vertexsize + 1)], dtype=NPY.int32, order="F"),
    [],
    CGK.IndexArray_ts,
    parent=zsr2,
)
CGL.newDataArray(zsr2, "var", value=NPY.ones(vertexsize, dtype="float64", order="F"))
zsr3 = CGU.newNode(
    "SubRegion3",
    NPY.array([2], dtype=NPY.int32, order="F"),
    [],
    CGK.ZoneSubRegion_ts,
    parent=z,
)
CGU.newNode(
    CGK.BCRegionName_s, CGU.setStringAsArray("BC"), [], CGK.Descriptor_ts, parent=zsr3
)
CGL.newDataArray(zsr3, "var", value=NPY.ones(ntris, dtype="float64", order="F"))
zsr5 = CGU.newNode(
    "SubRegion5",
    NPY.array([3], dtype=NPY.int32, order="F"),
    [],
    CGK.ZoneSubRegion_ts,
    parent=z,
)
CGU.newNode(
    CGK.PointRange_s,
    NPY.array([[1, cellsize]], dtype=NPY.int32, order="F"),
    [],
    CGK.IndexRange_ts,
    parent=zsr5,
)
CGL.newGridLocation(zsr5, value=CGK.CellCenter_s)
CGL.newDataArray(zsr5, "var", value=NPY.ones(cellsize, dtype="float64", order="F"))
zsr6 = CGU.newNode(
    "SubRegion6",
    NPY.array([3], dtype=NPY.int32, order="F"),
    [],
    CGK.ZoneSubRegion_ts,
    parent=z,
)
CGU.newNode(
    CGK.PointList_s,
    NPY.array([range(1, cellsize + 1)], dtype=NPY.int32, order="F"),
    [],
    CGK.IndexArray_ts,
    parent=zsr6,
)
CGL.newGridLocation(zsr6, value=CGK.CellCenter_s)
CGL.newDataArray(zsr6, "var", value=NPY.ones(cellsize, dtype="float64", order="F"))
TESTS.append((tag, T, diag))

#  -------------------------------------------------------------------------
tag = "zonesubregion bad node value"
diag = False
(T, b, z) = makeCorrectTree(vertexsize, cellsize, ntris)
zsr1 = CGU.newNode(
    "SubRegion1",
    NPY.array([4], dtype=NPY.int32, order="F"),
    [],
    CGK.ZoneSubRegion_ts,
    parent=z,
)
CGU.newNode(
    CGK.PointRange_s,
    NPY.array([[1, vertexsize]], dtype=NPY.int32, order="F"),
    [],
    CGK.IndexRange_ts,
    parent=zsr1,
)
TESTS.append((tag, T, diag))

#  -------------------------------------------------------------------------
tag = "zonesubregion bad Grid Location"
diag = False
(T, b, z) = makeCorrectTree(vertexsize, cellsize, ntris)
zsr1 = CGU.newNode(
    "SubRegion1",
    NPY.array([2], dtype=NPY.int32, order="F"),
    [],
    CGK.ZoneSubRegion_ts,
    parent=z,
)
CGU.newNode(
    CGK.PointRange_s,
    NPY.array([[1, cellsize]], dtype=NPY.int32, order="F"),
    [],
    CGK.IndexRange_ts,
    parent=zsr1,
)
CGL.newGridLocation(zsr1, value=CGK.CellCenter_s)
TESTS.append((tag, T, diag))

#  -------------------------------------------------------------------------
tag = "zonesubregion bad children"
diag = False
(T, b, z) = makeCorrectTree(vertexsize, cellsize, ntris)
zsr1 = CGU.newNode(
    "SubRegion1",
    NPY.array([3], dtype=NPY.int32, order="F"),
    [],
    CGK.ZoneSubRegion_ts,
    parent=z,
)
CGU.newNode(
    CGK.PointRange_s,
    NPY.array([[1, vertexsize]], dtype=NPY.int32, order="F"),
    [],
    CGK.IndexRange_ts,
    parent=zsr1,
)
CGU.newNode(
    CGK.PointList_s,
    NPY.array([range(1, vertexsize + 1)], dtype=NPY.int32, order="F"),
    [],
    CGK.IndexArray_ts,
    parent=zsr1,
)
CGU.newNode(
    CGK.BCRegionName_s, CGU.setStringAsArray("bc"), [], CGK.Descriptor_ts, parent=zsr1
)
TESTS.append((tag, T, diag))

#  -------------------------------------------------------------------------
tag = "zonesubregion bad regionname"
diag = False
(T, b, z) = makeCorrectTree(vertexsize, cellsize, ntris)
zsr1 = CGU.newNode(
    "SubRegion1",
    NPY.array([2], dtype=NPY.int32, order="F"),
    [],
    CGK.ZoneSubRegion_ts,
    parent=z,
)
CGU.newNode(
    CGK.BCRegionName_s, CGU.setStringAsArray("bc"), [], CGK.Descriptor_ts, parent=zsr1
)
zsr4 = CGU.newNode(
    "SubRegion4",
    NPY.array([2], dtype=NPY.int32, order="F"),
    [],
    CGK.ZoneSubRegion_ts,
    parent=z,
)
CGU.newNode(
    CGK.GridConnectivityRegionName_s,
    CGU.setStringAsArray("GC"),
    [],
    CGK.Descriptor_ts,
    parent=zsr4,
)
TESTS.append((tag, T, diag))

#  -------------------------------------------------------------------------
tag = "zonesubregion bad dataarray shapes"
diag = False
(T, b, z) = makeCorrectTree(vertexsize, cellsize, ntris)
zsr1 = CGU.newNode(
    "SubRegion1",
    NPY.array([3], dtype=NPY.int32, order="F"),
    [],
    CGK.ZoneSubRegion_ts,
    parent=z,
)
CGU.newNode(
    CGK.PointRange_s,
    NPY.array([[1, vertexsize]], dtype=NPY.int32, order="F"),
    [],
    CGK.IndexRange_ts,
    parent=zsr1,
)
CGL.newDataArray(
    zsr1, "var", value=NPY.ones(vertexsize + 1, dtype="float64", order="F")
)
zsr2 = CGU.newNode(
    "SubRegion2",
    NPY.array([3], dtype=NPY.int32, order="F"),
    [],
    CGK.ZoneSubRegion_ts,
    parent=z,
)
CGU.newNode(
    CGK.PointList_s,
    NPY.array([range(1, vertexsize + 1)], dtype=NPY.int32, order="F"),
    [],
    CGK.IndexArray_ts,
    parent=zsr2,
)
CGL.newDataArray(
    zsr2, "var", value=NPY.ones(vertexsize + 1, dtype="float64", order="F")
)
zsr3 = CGU.newNode(
    "SubRegion3",
    NPY.array([2], dtype=NPY.int32, order="F"),
    [],
    CGK.ZoneSubRegion_ts,
    parent=z,
)
CGU.newNode(
    CGK.BCRegionName_s, CGU.setStringAsArray("BC"), [], CGK.Descriptor_ts, parent=zsr3
)
CGL.newDataArray(zsr3, "var", value=NPY.ones(cellsize, dtype="float64", order="F"))
TESTS.append((tag, T, diag))

#  -------------------------------------------------------------------------
tag = "zonesubregion bad point list on cell and face subregions"
diag = False
(T, b, z) = makeCorrectTree(vertexsize, cellsize, ntris)
zsr6 = CGU.newNode(
    "SubRegion6",
    NPY.array([3], dtype=NPY.int32, order="F"),
    [],
    CGK.ZoneSubRegion_ts,
    parent=z,
)
CGU.newNode(
    CGK.PointList_s,
    NPY.array([range(1, cellsize + 2)], dtype=NPY.int32, order="F"),
    [],
    CGK.IndexArray_ts,
    parent=zsr6,
)
CGL.newGridLocation(zsr6, value=CGK.CellCenter_s)
zsr7 = CGU.newNode(
    "SubRegion7",
    NPY.array([3], dtype=NPY.int32, order="F"),
    [],
    CGK.ZoneSubRegion_ts,
    parent=z,
)
CGU.newNode(
    CGK.PointList_s,
    NPY.array([range(cellsize, cellsize + ntris)], dtype=NPY.int32, order="F"),
    [],
    CGK.IndexArray_ts,
    parent=zsr7,
)
CGL.newGridLocation(zsr7, value=CGK.FaceCenter_s)
TESTS.append((tag, T, diag))

#  -------------------------------------------------------------------------
tag = "zonesubregion bad point range on cell and face subregions"
diag = False
(T, b, z) = makeCorrectTree(vertexsize, cellsize, ntris)
zsr5 = CGU.newNode(
    "SubRegion5",
    NPY.array([3], dtype=NPY.int32, order="F"),
    [],
    CGK.ZoneSubRegion_ts,
    parent=z,
)
CGU.newNode(
    CGK.PointRange_s,
    NPY.array([[1, cellsize + 1]], dtype=NPY.int32, order="F"),
    [],
    CGK.IndexRange_ts,
    parent=zsr5,
)
CGL.newGridLocation(zsr5, value=CGK.CellCenter_s)
zsr8 = CGU.newNode(
    "SubRegion8",
    NPY.array([3], dtype=NPY.int32, order="F"),
    [],
    CGK.ZoneSubRegion_ts,
    parent=z,
)
CGU.newNode(
    CGK.PointRange_s,
    NPY.array([[cellsize, cellsize + ntris]], dtype=NPY.int32, order="F"),
    [],
    CGK.IndexRange_ts,
    parent=zsr8,
)
CGL.newGridLocation(zsr8, value=CGK.FaceCenter_s)
TESTS.append((tag, T, diag))



#  -------------------------------------------------------------------------
#  pyCGNS.VAL - Python package for CFD General Notation System - VALidater
#  See license.txt file in the root directory of this Python module source
#  -------------------------------------------------------------------------
#
import CGNS.PAT.cgnsutils as CGU
import CGNS.PAT.cgnstypes as CGT
import CGNS.PAT.cgnskeywords as CGK
import CGNS.PAT.cgnserrors as CGE
import CGNS.VAL.parse.messages as CGM

import inspect

genericmessages = {
    "G0001": (CGM.CHECK_FAIL, "CGNSLibraryVersion [%s] is too old wrt check level"),
    "G0002": (CGM.CHECK_FAIL, "CGNSLibraryVersion is incorrect"),
    "G0003": (CGM.CHECK_FAIL, "Name [%s] is not valid"),
    "G0004": (CGM.CHECK_FAIL, "Name [%s] is a duplicated child name"),
    "G0005": (CGM.CHECK_FAIL, "PANIC: Cannot find node with path [%s]"),
    "G0006": (CGM.CHECK_FAIL, "PANIC: Node data is not numpy.ndarray or None"),
    "G0007": (CGM.CHECK_FAIL, "PANIC: Node children is not a list"),
    "G0008": (CGM.CHECK_FAIL, "PANIC: Node name is not a string"),
    "G0009": (CGM.CHECK_FAIL, "PANIC: Node is not a list of 4 objects"),
    "G0010": (CGM.CHECK_FAIL, "PANIC: Node is empty list or None"),
    "G0011": (CGM.CHECK_FAIL, "PANIC: Node name is empty string"),
    "G0012": (CGM.CHECK_FAIL, "PANIC: Node name has forbidden chars"),
    "G0013": (CGM.CHECK_FAIL, "PANIC: Node name is . or .."),
    "G0014": (CGM.CHECK_FAIL, "PANIC: Node name is too long"),
    "G0015": (CGM.CHECK_FAIL, "Bad node value data type"),
    "S0001": (CGM.CHECK_FAIL, "Unknown SIDS type [%s]"),
    "S0002": (CGM.CHECK_FAIL, "SIDS type [%s] not allowed as child of [%s]"),
    "S0003": (CGM.CHECK_FAIL, "SIDS type [%s] not allowed for this node"),
    "S0004": (CGM.CHECK_FAIL, "DataType [%s] not allowed for this node"),
    "S0005": (CGM.CHECK_FAIL, "Node [%s] of type [%s] not allowed as child"),
    "S0006": (CGM.CHECK_FAIL, "Node [%s] of type [%s] allowed only once as child"),
    "S0007": (CGM.CHECK_FAIL, "Node [%s] of type [%s] is mandatory"),
    "S0008": (CGM.CHECK_FAIL, "Child name [%s] reserved for a type in [%s]"),
    "S0009": (CGM.CHECK_FAIL, "Bad node shape [%s]"),
    "S0010": (CGM.CHECK_FAIL, "Bad node value"),
}


class PathContext(dict):
    def __getitem__(self, path):
        if path not in self:
            return self.scope(path)
        return dict.__getitem__(self, path)

    def scope(self, path):
        if path in self:
            return dict.__getitem__(self, path)
        level = 1
        while True:
            apath = CGU.getPathAncestor(path, level)
            if apath in self:
                return dict.__getitem__(self, apath)
            if apath == "/":
                return None
            level += 1


class GenericContext(dict):
    def __getitem__(self, key):
        if key not in self:
            dict.__setitem__(self, key, PathContext())
        return dict.__getitem__(self, key)

    def __setitem__(self, key, value):
        d = self[key]
        d["/"] = value


class GenericParser(object):
    # --------------------------------------------------------------------
    def __init__(self, log=None):
        self._stop = False
        self.keywordlist = CGK.cgnsnames
        if log is None:
            self.log = CGM.DiagnosticLog()
        self.log.addMessages(genericmessages)
        self.context = GenericContext()
        self._trace = False

    # --------------------------------------------------------------------
    def dbg(self, msg, *args):
        if self._trace:
            print(
                "> %-24s" % msg,
            )
            if args:
                for a in args:
                    print(
                        a,
                    )
            print("")

    # --------------------------------------------------------------------
    def listDiagnostics(self):
        return self.log.listMessages()

    # --------------------------------------------------------------------
    def checkLeafStructure(self, T, path, node, parent):
        stt = CGM.CHECK_GOOD
        try:
            CGU.checkNode(node, dienow=True)
            CGU.checkNodeName(node, dienow=True)
            CGU.checkDuplicatedName(parent, node[0], dienow=True)
            CGU.checkNodeType(node, dienow=True)
            if node[1] is not None:
                CGU.checkArray(node[1], dienow=True)
        except CGE.cgnsException as v:
            if v.code == 1:
                stt = self.log.push(path, "G0010")
            if v.code == 2:
                stt = self.log.push(path, "G0009")
            if v.code == 3:
                stt = self.log.push(path, "G0008")
            if v.code == 4:
                stt = self.log.push(path, "G0007")
            if v.code == 5:
                stt = self.log.push(path, "G0006")
            if v.code == 22:
                stt = self.log.push(path, "G0008")
            if v.code == 23:
                stt = self.log.push(path, "G0011")
            if v.code == 24:
                stt = self.log.push(path, "G0012")
            if v.code == 25:
                stt = self.log.push(path, "G0014")
            if v.code == 29:
                stt = self.log.push(path, "G0013")
            if v.code == 31:
                stt = self.log.push(path, "G0011")
            if v.code == 32:
                stt = self.log.push(path, "G0011")
            if v.code == 111:
                stt = self.log.push(path, "G0015")
            if self._stop:
                raise CGE.cgnsException(-1)
        return stt

    # --------------------------------------------------------------------
    def checkLeaf(self, T, path, node):
        parent = CGU.getParentFromNode(T, node)
        status1 = self.checkSingleNode(T, path, node, parent)
        status2 = status1
        ntype = CGU.getTypeAsGrammarToken(node[3])
        if (len(node) == 4) and (ntype in self.methods):
            status2 = getattr(self, ntype)(*[path, node, parent, T, self.log])
        else:
            # if (ntype in CGK.cgnstypes): print '\nSKIP ',ntype
            pass
        status1 = CGM.getWorst(status1, status2)
        if self._stop and status1 == CGM.CHECK_FAIL:
            raise CGE.cgnsException(-1)
        return status1

    # --------------------------------------------------------------------
    def checkSingleNode(self, T, path, node, parent):
        stt = CGM.CHECK_GOOD
        if not CGU.checkNodeName(node):
            stt = self.log.push(path, "G0003", node[0])
        lchildren = CGU.childNames(parent)
        if lchildren:
            if node[0] in lchildren:
                lchildren.remove(node[0])
            if node[0] in lchildren:
                stt = self.log.push(path, "G0004", node[0])
        tlist = CGU.getNodeAllowedChildrenTypes(parent, node)
        if (CGU.getTypeAsGrammarToken(node[3]) not in tlist) and (
            node[3] != CGK.CGNSTree_ts
        ):
            if parent is not None:
                stt = self.log.push(path, "S0002", node[3], parent[3])
            else:
                stt = self.log.push(path, "S0003", node[3])
        dlist = CGU.getNodeAllowedDataTypes(node)
        dt = CGU.getValueDataType(node)
        if dt not in dlist:
            stt = self.log.push(path, "S0004", dt)
        if node[3] not in CGT.types.keys():
            stt = self.log.push(path, "S0001", node[3])
        else:
            stt = self.checkCardinalityOfChildren(T, path, node, parent)
            stt = self.checkReservedChildrenNames(T, path, node, parent)
        return stt

    # --------------------------------------------------------------------
    def checkTreeStructure(self, T, path=""):
        status = CGM.CHECK_GOOD
        status = self.checkLeafStructure(T, path, T, None)
        if status == CGM.CHECK_GOOD:
            path = path + "/" + T[0]
            for c in T[2]:
                status = self.checkTreeStructure(c, path)
        return status

    # --------------------------------------------------------------------
    def checkTree(self, T, trace=False, stop=False):
        self._stop = stop
        self._trace = trace
        status1 = CGM.CHECK_GOOD
        if self._trace:
            print("### Parsing node paths...")
        status1 = self.checkTreeStructure(T)
        if status1 != CGM.CHECK_GOOD:
            return status1
        paths = CGU.getPathFullTree(T, width=True)
        sz = len(paths) + 1
        ct = 1
        if not hasattr(self, "methods"):
            self.methods = []
            for m in inspect.getmembers(self):
                if (m[0][-2:] == "_t") or (m[0][-2:] == "_n") or (m[0][-3:] == "_ts"):
                    self.methods += [m[0]]
        for path in ["/"] + paths:
            if self._trace:
                print(
                    "### Check node [%.6d/%.6d]\r" % (ct, sz),
                )
            node = CGU.getNodeByPath(T, path)
            status2 = CGM.CHECK_GOOD
            if node is None:
                status2 = self.log.push(path, "G0005", path)
                if self._stop:
                    raise CGE.cgnsException(-1)
            if status2 == CGM.CHECK_GOOD:
                status2 = self.checkLeaf(T, path, node)
            status1 = status2
            ct += 1
        if self._trace:
            print("")
        return status1

    # --------------------------------------------------
    def checkCardinalityOfChildren(self, T, path, node, parent):
        stt = CGM.CHECK_GOOD
        for child in node[2]:
            card = CGT.types[node[3]].cardinality(child[3])
            if card == CGT.C_00:
                if path == "/":
                    cpath = "/%s" % (child[0])
                else:
                    cpath = "%s/%s" % (path, child[0])
                stt = self.log.push(path, "S0005", cpath, child[3])
            if card in [CGT.C_11, CGT.C_01]:
                if [c[3] for c in node[2]].count(child[3]) > 1:
                    stt = self.log.push(path, "S0006", child[0], child[3])
        for tchild in CGT.types[node[3]].children:
            card = CGT.types[node[3]].cardinality(tchild[0])
            if card in [CGT.C_11, CGT.C_1N]:
                if [c[3] for c in node[2]].count(tchild[0]) < 1:
                    stt = self.log.push(path, "S0007", tchild[1][0], tchild[0])
        return stt

    # --------------------------------------------------
    def checkReservedChildrenNames(self, T, path, node, parent):
        stt = CGM.CHECK_GOOD
        for child in node[2]:
            rt = CGT.types[node[3]].hasReservedNameType(child[0])
            if (rt != []) and (child[3] not in rt):
                srt = ""
                for s in rt:
                    srt = srt + "," + s
                stt = self.log.push(path, "S0008", child[0], srt[1:])
        return stt

    # --------------------------------------------------------------------
    def CGNSLibraryVersion_t(self, pth, node, parent, tree, log):
        stt = CGM.CHECK_OK
        try:
            version = int(node[1][0] * 1000)
            if version < 2400:
                stt = self.log.push(pth, "G0001", version)
            if version > 3400:
                stt = self.log.push(pth, "G0002")
        except Exception:
            stt = self.log.push(pth, "G0002")
        return stt


# --- last line



#  ---------------------------------------------------------------------------
#  pyCGNS - Python package for CFD General Notation System -
#  See license.txt file in the root directory of this Python module source
#  ---------------------------------------------------------------------------
#
import unittest


# --------------------------------------------------
# should not be documented with docstrings
class PATTestCase(unittest.TestCase):
    def setUp(self):
        self.T = None

    def eStr(self, code):
        import CGNS.PAT.cgnserrors as CGE

        return CGE.TAG + CGE.TAG_ERROR + "\[%.3d\].*$" % code

    def genTree(self):
        # should not self reference for test
        # then load tree
        import CGNS.PAT.test.disk

        self.T = CGNS.PAT.test.disk.T

    def test_00Module(self):
        pass

    def test_01Check(self):
        import CGNS.PAT.cgnsutils as CGU
        import CGNS.PAT.cgnserrors as CGE
        import CGNS.PAT.cgnskeywords as CGK

        self.assertFalse(CGU.checkName("/name"))
        self.assertFalse(CGU.checkNodeName(3))
        self.assertFalse(CGU.checkName(3))
        self.assertFalse(CGU.checkName(["name"]))
        self.assertFalse(CGU.checkName(""))
        self.assertFalse(CGU.checkName("x" * 33))
        self.assertFalse(CGU.checkName("."))
        self.assertFalse(CGU.checkName(".."))
        self.assertFalse(CGU.checkName("         "))
        self.assertTrue(CGU.checkName("name  test"))
        self.assertTrue(CGU.checkName('name"test"'))
        self.assertTrue(CGU.checkName("name\\test"))
        self.assertTrue(CGU.checkName(" name"))
        self.assertTrue(CGU.checkName("name"))
        self.assertTrue(CGU.checkName("x" * 32))
        self.assertTrue(CGU.checkNodeName(["test", None, [], "Zone_t"]))
        self.assertTrue(CGU.checkNodeName(["  test", None, [], "Zone_t"]))
        self.assertTrue(CGU.checkNodeName(["?test#320", None, [], "Zone_t"]))
        self.assertTrue(CGU.checkNodeName(["test   ", None, [], "Zone_t"]))
        self.assertFalse(CGU.checkName("name  test", strict=True))
        self.assertFalse(CGU.checkName("name\\test", strict=True))
        self.assertFalse(CGU.checkName('name"test"', strict=True))
        self.assertFalse(CGU.checkName(" name", strict=True))
        import string

        clist = list(string.ascii_letters + string.digits + string.punctuation + " ")
        clist.remove("/")
        for c in clist:
            self.assertTrue(CGU.checkName("_" + c))
        ex = CGE.cgnsNameError
        fn = CGU.checkName
        self.assertRaisesRegex(ex, self.eStr(2), CGU.checkNodeName, [], dienow=True)
        self.assertRaisesRegex(ex, self.eStr(22), fn, 3, dienow=True)
        self.assertRaisesRegex(ex, self.eStr(23), fn, "", dienow=True)
        self.assertRaisesRegex(ex, self.eStr(24), fn, "/", dienow=True)
        self.assertRaisesRegex(ex, self.eStr(25), fn, "a" * 33, dienow=True)
        self.assertRaisesRegex(ex, self.eStr(29), fn, ".", dienow=True)
        self.assertRaisesRegex(ex, self.eStr(31), fn, " " * 5, dienow=True)
        self.assertRaisesRegex(
            ex, self.eStr(32), fn, " a" * 5, dienow=True, strict=True
        )
        self.assertRaisesRegex(ex, self.eStr(33), fn, '"', dienow=True, strict=True)
        self.assertRaisesRegex(ex, self.eStr(34), fn, "a  a", dienow=True, strict=True)
        self.genTree()
        self.assertTrue(CGU.checkSameNode(self.T, self.T))
        self.assertFalse(CGU.checkSameNode(self.T, [None, None, [], None]))
        self.assertRaisesRegex(
            CGE.cgnsNodeError,
            self.eStr(30),
            CGU.checkSameNode,
            self.T,
            [None, None, [], None],
            dienow=True,
        )

    def test_02Path(self):
        import CGNS.PAT.cgnsutils as CGU
        import CGNS.PAT.cgnserrors as CGE
        import CGNS.PAT.cgnskeywords as CGK

        self.genTree()
        p = "/{Base#1}/{Zone-A}/ZoneBC"
        self.assertEqual(p, CGU.getPathNoRoot(p))
        self.assertEqual(p, CGU.getPathNoRoot("/CGNSTree" + p))
        self.assertEqual("/", CGU.getPathNoRoot("/CGNSTree"))
        self.assertEqual(CGU.getPathNormalize("/./{Base#1}/{Zone-A}/ZoneBC"), p)
        self.assertEqual(CGU.getPathNormalize("/./{Base#1}/{Zone-A}/ZoneBC//"), p)
        self.assertEqual(CGU.getPathToList(p), ["", "{Base#1}", "{Zone-A}", "ZoneBC"])
        self.assertEqual(CGU.getPathToList(p, True), ["{Base#1}", "{Zone-A}", "ZoneBC"])
        self.assertEqual(
            CGU.getPathAsTypes(self.T, p),
            [CGK.CGNSTree_ts, CGK.CGNSBase_ts, CGK.Zone_ts, CGK.ZoneBC_ts],
        )
        zone = ["Zone", None, [["ZoneBC", None, [], CGK.ZoneBC_ts]], CGK.Zone_ts]
        p = "/Zone/ZoneBC"
        self.assertEqual(
            CGU.getPathAsTypes(zone, p, legacy=False), [CGK.Zone_ts, CGK.ZoneBC_ts]
        )
        self.assertEqual(CGU.getPathAsTypes(zone, p, legacy=True), [CGK.ZoneBC_ts])

    def test_03NodeStructure(self):
        import CGNS.PAT.cgnsutils as CGU
        import CGNS.PAT.cgnserrors as CGE
        import CGNS.PAT.cgnskeywords as CGK

        self.assertEqual(["N1", None, [], "N_t"], CGU.nodeCreate("N1", None, [], "N_t"))
        p = CGU.nodeCreate("N1", None, [], "N_t")
        self.assertEqual(
            ["N1", None, [], "N_t"], CGU.nodeCreate("N1", None, [], "N_t", p)
        )
        self.assertEqual(["N1", None, [], "N_t"], p[2][0])
        n = CGU.nodeCreate("N2", None, [], "N_t", p)
        self.assertNotEqual(["N1", None, [], "N_t"], p[2][1])
        self.assertIs(n, p[2][1])
        self.assertEqual(["N2", None, [], "N_t"], CGU.nodeCopy(n))
        c = CGU.nodeCopy(n)
        self.assertIsNot(n, c)
        cname = "NCopy"
        c = CGU.nodeCopy(n, cname)
        self.assertEqual(c[0], cname)
        del n
        self.assertEqual(c[0], cname)

    def test_04NodeChildren(self):
        import CGNS.PAT.cgnsutils as CGU
        import CGNS.PAT.cgnslib as CGL
        import CGNS.PAT.cgnserrors as CGE
        import CGNS.PAT.cgnskeywords as CGK

        A = ["A", None, [], None]
        B = ["B", None, [], None]
        C = ["C", None, [], None]
        A[2] += [B]
        CGU.setAsChild(A, C)
        self.assertTrue(CGU.hasChildName(A, "B"))
        self.assertTrue(CGU.hasChildName(A, "C"))
        self.assertTrue(CGU.checkDuplicatedName(A, "D"))
        self.assertFalse(CGU.checkHasChildName(A, "D"))
        D = CGU.nodeCreate("D", None, [], None, parent=A)
        self.assertFalse(CGU.checkDuplicatedName(A, "D"))
        self.assertTrue(CGU.checkHasChildName(A, "D"))
        self.assertFalse(CGU.checkNodeType(C))
        d = {"None": None, "String": "string value", "Integer": 10, "Float": 1.4}
        for n, v in d.items():
            CGL.newDataArray(A, n, v)
        for name in d:
            self.assertTrue(CGU.hasChildName(A, name))

    def test_05NodeValue(self):
        import numpy
        import CGNS.PAT.cgnsutils as CGU
        import CGNS.PAT.cgnserrors as CGE
        import CGNS.PAT.cgnskeywords as CGK

        self.genTree()
        v = CGU.getValueByPath(self.T, "/CGNSTree/CGNSLibraryVersion")
        self.assertEqual(v, numpy.array(3.2))
        n = [
            "ZoneType",
            numpy.array(
                ["S", "t", "r", "u", "c", "t", "u", "r", "e", "d"], dtype="S", order="C"
            ),
            [],
            "ZoneType_t",
        ]
        self.assertTrue(CGU.stringValueMatches(n, "Structured"))
        # set*AsArray
        self.assertEqual(
            CGU.setStringAsArray("Structured").tostring().decode("ascii"), "Structured"
        )
        self.assertEqual(CGU.setIntegerAsArray(1), numpy.array(1, dtype="int32"))
        self.assertTrue(
            (
                CGU.setIntegerAsArray(1, 2, 3) == numpy.array([1, 2, 3], dtype="int32")
            ).all()
        )
        self.assertEqual(CGU.setLongAsArray(1), numpy.array(1, dtype="int64"))
        self.assertTrue(
            (CGU.setLongAsArray(1, 2, 3) == numpy.array([1, 2, 3], dtype="int64")).all()
        )
        self.assertEqual(CGU.setFloatAsArray(1), numpy.array(1, dtype="float32"))
        self.assertTrue(
            (
                CGU.setFloatAsArray(1, 2, 3) == numpy.array([1, 2, 3], dtype="float32")
            ).all()
        )
        self.assertEqual(CGU.setDoubleAsArray(1), numpy.array(1, dtype="float64"))
        self.assertTrue(
            (
                CGU.setDoubleAsArray(1, 2, 3) == numpy.array([1, 2, 3], dtype="float64")
            ).all()
        )
        self.assertTrue(
            (
                CGU.setDoubleAsArray(tuple(range(10, 1010, 10)))
                == numpy.array(tuple(range(10, 1010, 10)), dtype="float64")
            ).all()
        )
        n = ["ZoneType", None, [], "ZoneType_t"]
        # set*ByPath
        self.assertTrue(
            CGU.stringValueMatches(
                CGU.setStringByPath(n, "/", "Structured"), "Structured"
            )
        )
        self.assertEqual(
            CGU.setIntegerByPath(n, "/", 1)[1], numpy.array(1, dtype="int32")
        )
        self.assertTrue(
            numpy.array_equal(
                CGU.setIntegerByPath(n, "/", 1, 2, 3)[1],
                numpy.array([1, 2, 3], dtype="int32"),
            )
        )
        self.assertEqual(CGU.setLongByPath(n, "/", 1)[1], numpy.array(1, dtype="int64"))
        self.assertTrue(
            numpy.array_equal(
                CGU.setLongByPath(n, "/", 1, 2, 3)[1],
                numpy.array([1, 2, 3], dtype="int64"),
            )
        )
        self.assertEqual(
            CGU.setFloatByPath(n, "/", 1)[1], numpy.array(1, dtype="float32")
        )
        self.assertTrue(
            numpy.array_equal(
                CGU.setFloatByPath(n, "/", 1, 2, 3)[1],
                numpy.array([1, 2, 3], dtype="float32"),
            )
        )
        self.assertEqual(
            CGU.setDoubleByPath(n, "/", 1)[1], numpy.array(1, dtype="float64")
        )
        self.assertTrue(
            numpy.array_equal(
                CGU.setDoubleByPath(n, "/", 1, 2, 3)[1],
                numpy.array([1, 2, 3], dtype="float64"),
            )
        )
        self.assertTrue(
            numpy.array_equal(
                CGU.setDoubleByPath(n, "/", range(10, 1010, 10))[1],
                numpy.array([range(10, 1010, 10)], dtype="float64"),
            )
        )

    def test_06NodeType(self):
        import CGNS.PAT.cgnsutils as CGU
        import CGNS.PAT.cgnserrors as CGE
        import CGNS.PAT.cgnskeywords as CGK

        A = ["A", None, [], None]
        B = ["B", None, [], None]
        C = ["C", None, [], None]
        D = ["D", None, [], None]
        A[2] += [B]
        A[2] += [C]
        C[3] = "UnknownType_t"
        self.assertFalse(CGU.checkNodeType(C))
        self.assertFalse(CGU.checkNodeType(C, [CGK.CGNSBase_ts, CGK.Zone_ts]))
        self.assertFalse(CGU.checkNodeType(C, CGK.CGNSBase_ts))
        C[3] = CGK.CGNSBase_ts
        self.assertTrue(CGU.checkNodeType(C))
        self.assertTrue(CGU.checkNodeType(C, [CGK.CGNSBase_ts, CGK.Zone_ts]))
        self.assertTrue(
            CGU.checkNodeType(C, [CGK.Zone_ts, CGK.CGNSBase_ts, CGK.Zone_ts])
        )
        self.assertTrue(CGU.checkNodeType(C, CGK.CGNSBase_ts))

    def test_07NodeCompliance(self):
        import CGNS.PAT.cgnsutils as CGU
        import CGNS.PAT.cgnserrors as CGE
        import CGNS.PAT.cgnskeywords as CGK
        import numpy

        n = CGU.nodeCreate("Base", numpy.array([3, 3]), [], CGK.CGNSBase_ts)
        z = CGU.nodeCreate("ReferenceState", None, [], CGK.ReferenceState_ts, parent=n)
        self.assertTrue(CGU.checkNodeCompliant(n, None))
        self.assertTrue(CGU.checkNodeCompliant(z, None))
        self.assertTrue(CGU.checkNodeCompliant(z, n, dienow=True))
        self.assertFalse(CGU.checkNodeCompliant(None, n))

    def test_08NodeRetrieval(self):
        import CGNS.PAT.cgnsutils as CGU
        import CGNS.PAT.cgnserrors as CGE
        import CGNS.PAT.cgnskeywords as CGK

        self.genTree()
        p1 = "/CGNSTree/{Base#1}/{Zone-B}"
        self.assertEqual(CGU.getNodeByPath(self.T, p1)[0], "{Zone-B}")
        p2 = "/CGNSTree/.///{Base#1}/{Zone-B}/../{Zone-A}"
        self.assertEqual(CGU.getNodeByPath(self.T, p2)[0], "{Zone-A}")
        p3 = "/{Base#1}/{Zone-A}/ZoneBC"
        n3 = CGU.getNodeByPath(self.T, p3)
        self.assertEqual(n3[0], "ZoneBC")
        self.assertEqual(CGU.getPathFromNode(self.T, n3), p3)
        c1 = "{BC-2}"
        n31 = CGU.getNodeByPath(n3, c1)
        self.assertEqual(n31[0], "{BC-2}")
        self.assertEqual(CGU.getPathFromNode(n3, n31), "/" + c1)
        c2 = "./{BC-1}"
        self.assertEqual(CGU.getNodeByPath(n3, c2)[0], "{BC-1}")
        filter = "/.*/.*/Zone.*"
        v1 = ["/{Base#1}/{Zone-B}/ZoneBC", "/{Base#1}/{Zone-B}/ZoneGridConnectivity"]
        self.assertEqual(CGU.getPathByNameFilter(self.T, filter)[3:5], v1)
        filter = "/.*/.*/.*/GridConnectivity.*"
        v2 = "/{Base#1}/{Zone-D2}/ZoneGridConnectivity/{CT-D2-C}"
        self.assertEqual(CGU.getPathByTypeFilter(self.T, filter)[-2], v2)
        t = CGK.CGNSBase_ts
        b = CGU.getAncestorByType(self.T, n3, t)
        self.assertEqual(b[3], t)
        self.assertIsNone(CGU.getAncestorByType(self.T, b, CGK.Zone_ts))
        t = CGK.Zone_ts
        self.assertEqual(CGU.getAncestorByType(b, n3, t)[3], t)
        for p in CGU.getPathsByTypeSet(self.T, [CGK.BC_ts]):
            node = CGU.getNodeByPath(self.T, p)
            self.assertEqual(p, CGU.getPathFromNode(self.T, node))
        res = CGU.getPathsByTypeOrNameList(self.T, ["Zaza"])
        self.assertEqual(res, [])

    def test_09NodeDelete(self):
        import CGNS.PAT.cgnsutils as CGU
        import CGNS.PAT.cgnserrors as CGE
        import CGNS.PAT.cgnskeywords as CGK
        import numpy

        n = CGU.nodeCreate("Base", numpy.array([3, 3]), [], CGK.CGNSBase_ts)
        r = CGU.nodeCreate("ReferenceState", None, [], CGK.ReferenceState_ts, parent=n)
        d = CGU.nodeCreate("Data", numpy.array([3.14]), [], CGK.DataArray_ts, parent=r)
        self.assertIsNotNone(CGU.hasChildName(r, "Data"))
        CGU.nodeDelete(n, d)
        self.assertIsNone(CGU.hasChildName(r, "Data"))
        d = CGU.nodeCreate("DataZ", numpy.array([3.14]), [], CGK.DataArray_ts, parent=r)
        self.assertIsNotNone(CGU.hasChildName(r, "DataZ"))
        CGU.nodeDelete(r, d)
        self.assertIsNone(CGU.hasChildName(r, "DataZ"))
        CGU.nodeDelete(r, d)
        self.assertIsNone(CGU.hasChildName(r, "DataZ"))
        n = CGU.nodeCreate("Base", numpy.array([3, 3]), [], CGK.CGNSBase_ts)
        r = CGU.nodeCreate("ReferenceState", None, [], CGK.ReferenceState_ts, parent=n)
        d = CGU.nodeCreate("Data", numpy.array([3.14]), [], CGK.DataArray_ts, parent=r)
        self.assertIsNotNone(CGU.getNodeByPath(n, "/Base/ReferenceState/Data"))
        CGU.nodeDelete(n, d)
        self.assertIsNone(CGU.hasChildName(r, "Data"))
        n = CGU.nodeCreate("Base", numpy.array([3, 3]), [], CGK.CGNSBase_ts)
        r = CGU.nodeCreate("ReferenceState", None, [], CGK.ReferenceState_ts, parent=n)
        d = CGU.nodeCreate("Data", numpy.array([3.14]), [], CGK.DataArray_ts, parent=r)
        self.assertIsNotNone(CGU.getNodeByPath(n, "/Base/ReferenceState/Data"))
        CGU.removeChildByName(r, "Data")
        self.assertIsNone(CGU.hasChildName(r, "Data"))


# ---
print("-" * 70 + "\nCGNS.PAT test suite")
suite = unittest.TestLoader().loadTestsFromTestCase(PATTestCase)
unittest.TextTestRunner(verbosity=2).run(suite)

# --- last line



#  -------------------------------------------------------------------------
#  pyCGNS - Python package for CFD General Notation System -
#  See license.txt file in the root directory of this Python module source
#  -------------------------------------------------------------------------
#
import CGNS.PAT.cgnslib as CGL
import CGNS.PAT.cgnsutils as CGU
import CGNS.PAT.cgnskeywords as CGK
import numpy as NPY

TESTS = []

#  -------------------------------------------------------------------------
tag = "zone structured"
diag = True
T = CGL.newCGNSTree()
b = CGL.newBase(T, "{Base}", 3, 3)
z = CGL.newZone(b, "{Zone}", NPY.array([[5, 4, 0], [7, 6, 0], [9, 8, 0]], order="F"))
g = CGL.newGridCoordinates(z, "GridCoordinates")
d = CGL.newDataArray(
    g, CGK.CoordinateX_s, NPY.ones((5, 7, 9), dtype="float64", order="F")
)
d = CGL.newDataArray(
    g, CGK.CoordinateY_s, NPY.ones((5, 7, 9), dtype="float64", order="F")
)
d = CGL.newDataArray(
    g, CGK.CoordinateZ_s, NPY.ones((5, 7, 9), dtype="float64", order="F")
)
TESTS.append((tag, T, diag))

#  -------------------------------------------------------------------------
tag = "zone bad zonetype"
diag = False
T = CGL.newCGNSTree()
b = CGL.newBase(T, "{Base}", 3, 3)
z = CGL.newZone(b, "{Zone}", NPY.array([[5, 4, 0], [7, 6, 0], [9, 8, 0]], order="F"))
g = CGL.newGridCoordinates(z, "GridCoordinates")
d = CGL.newDataArray(
    g, CGK.CoordinateX_s, NPY.ones((5, 7, 9), dtype="float64", order="F")
)
d = CGL.newDataArray(
    g, CGK.CoordinateY_s, NPY.ones((5, 7, 9), dtype="float64", order="F")
)
d = CGL.newDataArray(
    g, CGK.CoordinateZ_s, NPY.ones((5, 7, 9), dtype="float64", order="F")
)
zt = CGU.hasChildName(z, CGK.ZoneType_s)
zt[1] = CGU.setStringAsArray("Untruscutred")
TESTS.append((tag, T, diag))

#  -------------------------------------------------------------------------
tag = "zone bad zone size"
diag = False
T = CGL.newCGNSTree()
b = CGL.newBase(T, "{Base}", 3, 3)
z = CGL.newZone(b, "{Zone}", NPY.array([[5, 4, 0], [7, 6, 0], [9, 8, 0]], order="F"))
g = CGL.newGridCoordinates(z, "GridCoordinates")
d = CGL.newDataArray(g, CGK.CoordinateX_s, NPY.ones(5, dtype="float64", order="F"))
d = CGL.newDataArray(g, CGK.CoordinateY_s, NPY.ones(5, dtype="float64", order="F"))
d = CGL.newDataArray(g, CGK.CoordinateZ_s, NPY.ones(5, dtype="float64", order="F"))
zt = CGU.hasChildName(z, CGK.ZoneType_s)
zt[1] = CGU.setStringAsArray(CGK.Unstructured_s)
tetras = CGL.newElements(
    z,
    "TETRAS",
    CGK.TETRA_4_s,
    NPY.array([[1, 4]], "i", order="F"),
    NPY.ones((4 * 4), dtype="int32"),
)
TESTS.append((tag, T, diag))

#  -------------------------------------------------------------------------
tag = "zone unstructured bad dims"
diag = False
T = CGL.newCGNSTree()
b = CGL.newBase(T, "{Base}", 3, 3)
z = CGL.newZone(b, "{Zone}", NPY.array([[4, 5, 0]], order="F"))
g = CGL.newGridCoordinates(z, "GridCoordinates")
d = CGL.newDataArray(g, CGK.CoordinateX_s, NPY.ones(4, dtype="float64", order="F"))
d = CGL.newDataArray(g, CGK.CoordinateY_s, NPY.ones(4, dtype="float64", order="F"))
d = CGL.newDataArray(g, CGK.CoordinateZ_s, NPY.ones(4, dtype="float64", order="F"))
zt = CGU.hasChildName(z, CGK.ZoneType_s)
zt[1] = CGU.setStringAsArray(CGK.Unstructured_s)
tetras = CGL.newElements(
    z,
    "TETRAS",
    CGK.TETRA_4_s,
    NPY.array([[1, 5]], "i", order="F"),
    NPY.ones((5 * 4), dtype="int32"),
)
TESTS.append((tag, T, diag))

#  -------------------------------------------------------------------------
tag = "zone structured bad dims"
diag = False
T = CGL.newCGNSTree()
b = CGL.newBase(T, "{Base}", 3, 3)
z = CGL.newZone(b, "{Zone}", NPY.array([[5, 4, 0], [7, 7, 0], [9, 8, 0]], order="F"))
g = CGL.newGridCoordinates(z, "GridCoordinates")
d = CGL.newDataArray(
    g, CGK.CoordinateX_s, NPY.ones((5, 7, 9), dtype="float64", order="F")
)
d = CGL.newDataArray(
    g, CGK.CoordinateY_s, NPY.ones((5, 7, 9), dtype="float64", order="F")
)
d = CGL.newDataArray(
    g, CGK.CoordinateZ_s, NPY.ones((5, 7, 9), dtype="float64", order="F")
)
TESTS.append((tag, T, diag))

#  -------------------------------------------------------------------------
tag = "zone structured bad numpy array order"
diag = False
T = CGL.newCGNSTree()
b = CGL.newBase(T, "{Base}", 3, 3)
z = CGL.newZone(b, "{Zone}", NPY.array([[5, 7, 9], [4, 6, 8], [0, 0, 0]], order="F"))
g = CGL.newGridCoordinates(z, "GridCoordinates")
d = CGL.newDataArray(
    g, CGK.CoordinateX_s, NPY.ones((5, 4, 0), dtype="float64", order="F")
)
d = CGL.newDataArray(
    g, CGK.CoordinateY_s, NPY.ones((5, 4, 0), dtype="float64", order="F")
)
d = CGL.newDataArray(
    g, CGK.CoordinateZ_s, NPY.ones((5, 4, 0), dtype="float64", order="F")
)
TESTS.append((tag, T, diag))

#  -------------------------------------------------------------------------
tag = "zone unstructured no elements"
diag = False


def makeUnstTree(vertexsize, cellsize):
    T = CGL.newCGNSTree()
    b = CGL.newBase(T, "Base", 3, 3)
    s = NPY.array([[vertexsize, cellsize, 0]], dtype="int32", order="F")
    z = CGL.newZone(b, "Zone", s, CGK.Unstructured_s)
    g = CGL.newGridCoordinates(z, "GridCoordinates")
    d = CGL.newDataArray(
        g, CGK.CoordinateX_s, NPY.ones((vertexsize), dtype="float64", order="F")
    )
    d = CGL.newDataArray(
        g, CGK.CoordinateY_s, NPY.ones((vertexsize), dtype="float64", order="F")
    )
    d = CGL.newDataArray(
        g, CGK.CoordinateZ_s, NPY.ones((vertexsize), dtype="float64", order="F")
    )
    return (T, b, z)


vertexsize = 20
cellsize = 7
(T, b, z) = makeUnstTree(vertexsize, cellsize)
TESTS.append((tag, T, diag))

#  -------------------------------------------------------------------------
tag = "zone unstructured correct ElementRange combination"
diag = True
(T, b, z) = makeUnstTree(vertexsize, cellsize)
tetras = CGL.newElements(
    z,
    "TETRAS",
    CGK.TETRA_4_s,
    NPY.array([[1, 3]], "i", order="F"),
    NPY.ones((3 * 4), dtype="int32"),
)
hexas = CGL.newElements(
    z,
    "HEXAS",
    CGK.HEXA_8_s,
    NPY.array([[4, cellsize]], "i", order="F"),
    NPY.ones((4 * 8), dtype="int32"),
)
element = CGL.newElements(
    z,
    "NGON",
    CGK.NGON_n_s,
    NPY.array([[cellsize + 1, cellsize + 6]], "i", order="F"),
    NPY.array(
        [9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9],
        dtype="int32",
        order="F",
    ),
    NPY.array([0, 4, 9, 12, 17, 20, 24], dtype="int32", order="F"),
)
hexas2 = CGL.newElements(
    z,
    "HEXAS2",
    CGK.HEXA_8_s,
    NPY.array([[cellsize + 7, cellsize + 8]], "i", order="F"),
    NPY.ones((2 * 8), dtype="int32"),
)
TESTS.append((tag, T, diag))

#  -------------------------------------------------------------------------
tag = "zone unstructured bad ElementRange combination #1"
diag = False
(T, b, z) = makeUnstTree(vertexsize, cellsize)
tetras = CGL.newElements(
    z,
    "TETRAS",
    CGK.TETRA_4_s,
    NPY.array([[1, 3]], "i", order="F"),
    NPY.ones((3 * 4), dtype="int32"),
)
hexas = CGL.newElements(
    z,
    "HEXAS",
    CGK.HEXA_8_s,
    NPY.array([[3, 7]], "i", order="F"),
    NPY.ones((5 * 8), dtype="int32"),
)  # Bad combination of ElementRange
TESTS.append((tag, T, diag))

#  -------------------------------------------------------------------------
tag = "zone unstructured bad ElementRange combination #2"
diag = False
(T, b, z) = makeUnstTree(vertexsize, cellsize)
tetras = CGL.newElements(
    z,
    "TETRAS",
    CGK.TETRA_4_s,
    NPY.array([[1, cellsize]], "i", order="F"),
    NPY.ones((cellsize * 4), dtype="int32"),
)
element = CGL.newElements(
    z,
    "NGON",
    CGK.NGON_n_s,
    NPY.array([[cellsize + 1 + 1, cellsize + 7 + 1]], "i", order="F"),
    NPY.array(
        [
            9,
            9,
            9,
            9,
            9,
            9,
            9,
            9,
            9,
            9,
            9,
            9,
            9,
            9,
            9,
            9,
            9,
            9,
            9,
            9,
            9,
            9,
            9,
            9,
            9,
            9,
            9,
            9,
            9,
        ],
        dtype="int32",
        order="F",
    ),
    NPY.array([0, 4, 9, 12, 17, 20, 25, 29], dtype="int32", order="F"),
)  # should be cellsize+1,cellsize+7
TESTS.append((tag, T, diag))

#  -------------------------------------------------------------------------
tag = "zone unstructured bad ElementRange combination #3"
diag = False
(T, b, z) = makeUnstTree(vertexsize, cellsize)
tetras = CGL.newElements(
    z,
    "TETRAS",
    CGK.TETRA_4_s,
    NPY.ones((cellsize * 4), dtype="int32"),
    NPY.array([[1, cellsize]], "i", order="F"),
)
element = CGL.newElements(
    z,
    "NGON",
    CGK.NGON_n_s,
    NPY.array([[cellsize, cellsize + 7 - 1]], "i", order="F"),
    NPY.array(
        [
            9,
            9,
            9,
            9,
            9,
            9,
            9,
            9,
            9,
            9,
            9,
            9,
            9,
            9,
            9,
            9,
            9,
            9,
            9,
            9,
            9,
            9,
            9,
            9,
            9,
            9,
            9,
            9,
            9,
        ],
        dtype="int32",
        order="F",
    ),
    NPY.array([0, 4, 9, 12, 17, 20, 25, 29], dtype="int32", order="F"),
)  # should be cellsize+1,cellsize+7
TESTS.append((tag, T, diag))

#  -------------------------------------------------------------------------
tag = "zone unstructured NFACE without NGON"
diag = False
(T, b, z) = makeUnstTree(vertexsize, cellsize)
tetras = CGL.newElements(
    z,
    "TETRAS",
    CGK.TETRA_4_s,
    NPY.array([[1, cellsize]], "i", order="F"),
    NPY.ones((cellsize * 4), dtype="int32"),
)
element = CGL.newElements(
    z,
    "NFACE",
    CGK.NFACE_n_s,
    NPY.array([[cellsize + 1, cellsize + 5]], "i", order="F"),
    NPY.array(
        [9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9, 9],
        dtype="int32",
        order="F",
    ),
    NPY.array([0, 4, 9, 12, 17, 21], dtype="int32", order="F"),
)
TESTS.append((tag, T, diag))

#  -------------------------------------------------------------------------
tag = "zone structured full BC and GridConnect"
diag = True


def makeStTree():
    T = CGL.newCGNSTree()
    b = CGL.newBase(T, "{Base}", 3, 3)
    z1 = CGL.newZone(
        b, "{Zone1}", NPY.array([[5, 4, 0], [7, 6, 0], [9, 8, 0]], order="F")
    )
    g = CGL.newGridCoordinates(z1, "GridCoordinates")
    d = CGL.newDataArray(
        g, CGK.CoordinateX_s, NPY.ones((5, 7, 9), dtype="float64", order="F")
    )
    d = CGL.newDataArray(
        g, CGK.CoordinateY_s, NPY.ones((5, 7, 9), dtype="float64", order="F")
    )
    d = CGL.newDataArray(
        g, CGK.CoordinateZ_s, NPY.ones((5, 7, 9), dtype="float64", order="F")
    )
    z2 = CGU.copyNode(z1, "{Zone2}")
    b[2].append(z2)
    zgc = CGL.newZoneGridConnectivity(z1)
    gc = CGL.newGridConnectivity1to1(
        zgc,
        "join1_2",
        "{Zone2}",
        NPY.array([[1, 1], [1, 4], [1, 9]]),
        NPY.array([[5, 5], [3, 7], [1, 9]]),
        NPY.array([-1, +2, +3]),
    )
    zgc = CGL.newZoneGridConnectivity(z2)
    gc = CGL.newGridConnectivity1to1(
        zgc,
        "join2_1",
        "{Zone1}",
        NPY.array([[5, 5], [3, 7], [1, 9]]),
        NPY.array([[1, 1], [1, 4], [1, 9]]),
        NPY.array([-1, +2, +3]),
    )
    zbc = CGL.newZoneBC(z1)
    n = CGL.newBoundary(
        zbc,
        "{BC1_1}",
        [[5, 5], [1, 7], [1, 9]],
        btype=CGK.Null_s,
        family=None,
        pttype=CGK.PointRange_s,
    )
    g = CGL.newGridLocation(n, value=CGK.FaceCenter_s)
    n = CGL.newBoundary(
        zbc,
        "{BC1_2}",
        [[1, 5], [1, 1], [1, 9]],
        btype=CGK.Null_s,
        family=None,
        pttype=CGK.PointRange_s,
    )
    g = CGL.newGridLocation(n, value=CGK.FaceCenter_s)
    n = CGL.newBoundary(
        zbc,
        "{BC1_3}",
        [[1, 5], [7, 7], [1, 9]],
        btype=CGK.Null_s,
        family=None,
        pttype=CGK.PointRange_s,
    )
    g = CGL.newGridLocation(n, value=CGK.FaceCenter_s)
    n = CGL.newBoundary(
        zbc,
        "{BC1_4}",
        [[1, 5], [1, 7], [1, 1]],
        btype=CGK.Null_s,
        family=None,
        pttype=CGK.PointRange_s,
    )
    g = CGL.newGridLocation(n, value=CGK.FaceCenter_s)
    n = CGL.newBoundary(
        zbc,
        "{BC1_5}",
        [[1, 5], [1, 7], [9, 9]],
        btype=CGK.Null_s,
        family=None,
        pttype=CGK.PointRange_s,
    )
    g = CGL.newGridLocation(n, value=CGK.FaceCenter_s)
    n = CGL.newBoundary(
        zbc,
        "{BC1_6}",
        [[1, 1], [4, 7], [1, 9]],
        btype=CGK.Null_s,
        family=None,
        pttype=CGK.PointRange_s,
    )
    g = CGL.newGridLocation(n, value=CGK.FaceCenter_s)
    zbc = CGL.newZoneBC(z2)
    n = CGL.newBoundary(
        zbc,
        "{BC2_1}",
        [[1, 1], [1, 7], [1, 9]],
        btype=CGK.Null_s,
        family=None,
        pttype=CGK.PointRange_s,
    )
    g = CGL.newGridLocation(n, value=CGK.FaceCenter_s)
    n = CGL.newBoundary(
        zbc,
        "{BC2_2}",
        [[1, 5], [1, 1], [1, 9]],
        btype=CGK.Null_s,
        family=None,
        pttype=CGK.PointRange_s,
    )
    g = CGL.newGridLocation(n, value=CGK.FaceCenter_s)
    n = CGL.newBoundary(
        zbc,
        "{BC2_3}",
        [[1, 5], [7, 7], [1, 9]],
        btype=CGK.Null_s,
        family=None,
        pttype=CGK.PointRange_s,
    )
    g = CGL.newGridLocation(n, value=CGK.FaceCenter_s)
    n = CGL.newBoundary(
        zbc,
        "{BC2_4}",
        [[1, 5], [1, 7], [1, 1]],
        btype=CGK.Null_s,
        family=None,
        pttype=CGK.PointRange_s,
    )
    g = CGL.newGridLocation(n, value=CGK.FaceCenter_s)
    n = CGL.newBoundary(
        zbc,
        "{BC2_5}",
        [[1, 5], [1, 7], [9, 9]],
        btype=CGK.Null_s,
        family=None,
        pttype=CGK.PointRange_s,
    )
    g = CGL.newGridLocation(n, value=CGK.FaceCenter_s)
    n = CGL.newBoundary(
        zbc,
        "{BC2_6}",
        [[5, 5], [1, 3], [1, 9]],
        btype=CGK.Null_s,
        family=None,
        pttype=CGK.PointRange_s,
    )
    g = CGL.newGridLocation(n, value=CGK.FaceCenter_s)
    z = [z1, z2]
    return (T, b, z)


(T, b, z) = makeStTree()
TESTS.append((tag, T, diag))

#  -------------------------------------------------------------------------
tag = "zone structured uncomplete BC and GridConnect (warning)"
diag = False
(T, b, z) = makeStTree()
pth = CGU.getAllNodesByTypeOrNameList(z[0], ["Zone_t", "ZoneBC_t", "{BC1_1}"])[0]
CGU.removeChildByName(CGU.getNodeByPath(z[0], CGU.getPathAncestor(pth)), "{BC1_1}")
pth = CGU.getAllNodesByTypeOrNameList(
    z[1], ["Zone_t", "ZoneGridConnectivity_t", "join2_1"]
)[0]
CGU.removeChildByName(CGU.getNodeByPath(z[1], CGU.getPathAncestor(pth)), "join2_1")
TESTS.append((tag, T, diag))

#  -------------------------------------------------------------------------
tag = "zone structured doubly defined BC and GridConnect (warning)"
diag = False
(T, b, z) = makeStTree()
pth = CGU.getAllNodesByTypeOrNameList(z[0], ["Zone_t", "ZoneBC_t"])[0]
zbc = CGU.getNodeByPath(z[0], pth)
n = CGL.newBoundary(
    zbc,
    "{BC1_1b}",
    [[5, 5], [1, 2], [1, 2]],
    btype=CGK.Null_s,
    family=None,
    pttype=CGK.PointRange_s,
)
g = CGL.newGridLocation(n, value=CGK.FaceCenter_s)
pth = CGU.getAllNodesByTypeOrNameList(z[1], ["Zone_t", "ZoneGridConnectivity_t"])[0]
zgc = CGU.getNodeByPath(z[1], pth)
gc = CGL.newGridConnectivity1to1(
    zgc,
    "join2_1b",
    "{Zone1}",
    NPY.array([[1, 1], [1, 2], [1, 2]]),
    NPY.array([[1, 1], [1, 4], [1, 9]]),
    NPY.array([-1, +2, +3]),
)
TESTS.append((tag, T, diag))
